package edu.emory.mathcs;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import edu.emory.mathcs.measures.Measure;
import edu.emory.mathcs.measures.MeasuresFactory;
import edu.emory.mathcs.measures.privacy.DeltaGrowthGroup;
import edu.emory.mathcs.measures.utility.Entropy;
import edu.emory.mathcs.privacy.AndCheckers;
import edu.emory.mathcs.privacy.Entropy_L_diversity;
import edu.emory.mathcs.privacy.K_anonymity;
import edu.emory.mathcs.privacy.M_privacy;
import edu.emory.mathcs.privacy.PrivacyChecker;
import edu.emory.mathcs.privacy.T_closeness;

public class CombinationsCounter
{
	final private TuplesGroup tuples;
	final private Map<Site, List<Tuple>> site2tuples;
	final private Map<Site, Integer> site2numTuples;
	final private PrivacyChecker checker;
//	final int actualNumSites;
	private boolean withAdaptiveOrdering;
	
//	public CombinationsCounter(List<Integer> tuplesDistr, int numPrivateTuples){
//		this.name2tuplesDistribution = new HashMap<String, Integer>();
//		int counter = 1;
//		for (int freq : tuplesDistr){
//			this.name2tuplesDistribution.put("" + counter++, freq);
//		}
//		this.tuples = tuples;
//		//this.numPrivateTuples = numPrivateTuples;
//		this.checker = new K_anonymity(numPrivateTuples);
//	}

	public CombinationsCounter(TuplesGroup group, PrivacyChecker checker){
		this.tuples = group;
		this.checker = checker;
		this.site2tuples = this.tuples.getTuplesSplittedToSites();
		this.site2numTuples = this.tuples.getSiteFrequencies();
		this.withAdaptiveOrdering = true;
//		this.actualNumSites = this.site2numTuples.size();
		
//		Set<Site> sites = this.tuples.getSites();
//		for (Site s: sites){
//			if (!this.site2tuples.containsKey(s))
//				this.site2tuples.put(s, new ArrayList<Tuple>());
//			if (!this.site2numTuples.containsKey(s))
//				this.site2numTuples.put(s, 0);
//		}
		
	}

	public TuplesGroup getTuples(Collection<Site> sites){
		List<Tuple> tuples = new ArrayList<Tuple>();
		
		for (Site s: sites)
			tuples.addAll(this.site2tuples.get(s));
		
		return new TuplesGroup(tuples);
	}
	
	public double getPrivacyLevel(Collection<Tuple> tuples) {
		return this.checker.getPrivacyFitness(new TuplesGroup(tuples));
	}
	
	public void setAdaptiveOrdering(boolean value){
		this.withAdaptiveOrdering = value;
	}
	
	public HashMap<Integer, Integer> simulateCheckingWithoutSwitch(int m, boolean attackers){
		int numSites = this.site2numTuples.size();
		return simulateChecking(m, new int[] {numSites - 1, m}, false, attackers);
	}
	
	public HashMap<Integer, Integer> simulateCheckingWithSwitch(int m, boolean attackers){
		int numSites = this.site2numTuples.size();
		return simulateChecking(m, new int[] {numSites - 1, m}, true, attackers);
	}

	private List<Site> getSitesSortedByAPrivacyFitness() {
		List<Site> sortedSites = new ArrayList<Site>(this.site2tuples.size());
		Map<Site, Double> site2privacyFitness = new HashMap<Site, Double>();
		Set<Site> sites = this.site2tuples.keySet();
		
		if (!this.withAdaptiveOrdering){
			sortedSites.addAll(sites);
			Collections.shuffle(sortedSites);
			return sortedSites;
		}
		
		for (Site s: sites){
			TuplesGroup group = new TuplesGroup(this.site2tuples.get(s));
			site2privacyFitness.put(s, this.checker.getPrivacyFitness(group));
		}
		sortedSites.addAll(sites);
		final Map<Site, Double> finalSite2privacyLevel = new HashMap<Site, Double>(site2privacyFitness);
		Collections.sort(sortedSites, new Comparator<Site>() {
			@Override
			public int compare(Site site0, Site site1)
			{
				return (int)Math.signum(finalSite2privacyLevel.get(site0) - finalSite2privacyLevel.get(site1));
			}
		});
		
		return sortedSites;
	}
	
	private HashMap<Integer, Integer> simulateChecking(int m, int[] possibleAttackers, boolean withSwitch,
			boolean attackers){
		HashMap<Integer, Integer> groupSize2Freq = new HashMap<Integer, Integer>();
		
		// weak attackers will be generated first
		List<Site> sites = this.getSitesSortedByAPrivacyFitness();
		CombinationsGenerator<Site> generator = new CombinationsGenerator<Site>(sites);
		CombinationsGenerator<Site> checker = new CombinationsGenerator<Site>(sites);
		List<Site> revSites = new ArrayList<Site>(sites);
		Collections.reverse(revSites);
		CombinationsGenerator<Site> revGenerator = new CombinationsGenerator<Site>(revSites);
		checker.setCardinality(m);
		if (this.tuples.getNumTuples() > 0 && !this.checker.isAnonymized(this.tuples))
			return groupSize2Freq;
		int numAllTuples = sum(this.site2numTuples.keySet());//sum(this.name2tuplesDistribution.keySet());

		// checking n-1 and then m
		for (int cardinality : possibleAttackers) {
			if (cardinality == this.site2numTuples.size()/*this.name2tuplesDistribution.size()*/) {
				if (attackers)
					groupSize2Freq.put(numAllTuples, 1);
				else
					groupSize2Freq.put(0, 1);
				continue;
			}

			if (cardinality <= 0)
				continue;
			
			generator.setCardinality(cardinality);
			revGenerator.setCardinality(cardinality);
			while ((generator.hasNext() || revGenerator.hasNext()) 
					&& !checker.areThisAndAllNextDecided()) {
				// for cardinality >= m use weaker attackers first!
				Set<Site> attackingSites = (cardinality >= m ? generator.next() : revGenerator.next());
				
				if (attackingSites == null)
					throw new IllegalArgumentException("Generators are not synchronized.");
				
				int numTuples = sum(attackingSites);
				if (!attackers)
					numTuples = numAllTuples - numTuples;
				if (groupSize2Freq.containsKey(numTuples) == false) {
					groupSize2Freq.put(numTuples, 0);
				}
				int freq = groupSize2Freq.get(numTuples);
				groupSize2Freq.put(numTuples, freq + 1);
				
				Set<Site> defenders = new HashSet<Site>(this.site2numTuples.keySet());
				defenders.removeAll(attackingSites);
				if (defenders.size() == 0)
					continue;
				if (!this.checker.isAnonymized(this.getTuples(defenders))){
					generator.forbidThisAndAllSuper(attackingSites);
					revGenerator.forbidThisAndAllSuper(attackingSites);
					if (cardinality <= m)
						return groupSize2Freq;
					else
						checker.forbidThisAndAllSuper(attackingSites);
					if (withSwitch)
						break;
				}
				else {
					generator.forbidThisAndAllSub(attackingSites);
					revGenerator.forbidThisAndAllSub(attackingSites);
					checker.forbidThisAndAllSub(attackingSites);
				}
			}
			
			if (checker.areThisAndAllNextDecided())
				break;
		}
		
		return groupSize2Freq;
	}

	private void reportPrivacyCheck(HashMap<Integer, Integer> groupSize2Freq, int groupSize) {
		if (groupSize2Freq.containsKey(groupSize) == false)
			groupSize2Freq.put(groupSize, 0);
		groupSize2Freq.put(groupSize, groupSize2Freq.get(groupSize) + 1);
	}
	
	public HashMap<Integer, Integer> simulateCheckingWithBinaryJumping(int m, boolean attackers){
		HashMap<Integer, Integer> groupSize2Freq = new HashMap<Integer, Integer>();
		
		List<Site> sites = this.getSitesSortedByAPrivacyFitness();
		CombinationsGenerator<Site> cg = new CombinationsGenerator<Site>(sites);
		CombinationsGenerator<Site> mCombinationsChecker = new CombinationsGenerator<Site>(sites);
		mCombinationsChecker.setCardinality(m);
		int maxNumAttackers = this.site2numTuples.size();
		int numAttackers = maxNumAttackers;
		int allTuples = this.tuples.getNumTuples();

		while (!mCombinationsChecker.areThisAndAllNextDecided()){
			int numMoreAttackers = numAttackers;
			int numLessAttackers = m;
			cg.setCardinality(m);
			if (!cg.hasNext())
				continue;
			Set<Site> lessAttackers = cg.next();

			// report check for less attackers
			int toReport = sum(lessAttackers);
			if (!attackers)
				toReport = allTuples - toReport;
			this.reportPrivacyCheck(groupSize2Freq, toReport);
			
			// Do privacy check for less attackers
			Set<Site> defenders = new HashSet<Site>(sites);
			defenders.removeAll(lessAttackers);
			TuplesGroup defendedGroup = this.getTuples(defenders);
			boolean isAnonymizedForLessAttackers = this.checker.isAnonymized(defendedGroup);
			if (defenders.size() > 0 
					&& !isAnonymizedForLessAttackers && lessAttackers.size() <= m)
				return groupSize2Freq;

			// prepare set of more attackers
			Set<Site> moreAttackers = cg.nextSuperSet(lessAttackers, numMoreAttackers - numLessAttackers);
			while (moreAttackers == null && numAttackers > m) {
				numMoreAttackers = --numAttackers;
				cg.setCardinality(numMoreAttackers);
				moreAttackers = cg.next();
				if (moreAttackers.equals(lessAttackers)){
					moreAttackers = cg.next();
				}
			}
			if (moreAttackers == null) {
				int cardinality = cg.getCardinality();
				cg.setCardinality(cardinality);
				moreAttackers = cg.next();
			}
			
			// Report privacy check for more attackers
			toReport = sum(moreAttackers);
			if (!attackers)
				toReport = allTuples - toReport;
			this.reportPrivacyCheck(groupSize2Freq, toReport);

			// Do privacy check for more attackers
			defenders = new HashSet<Site>(sites);
			defenders.removeAll(moreAttackers);
			defendedGroup = this.getTuples(defenders);
			boolean isAnonymizedForMoreAttackers = this.checker.isAnonymized(defendedGroup);
			if (defenders.size() > 0 
					&& !isAnonymizedForMoreAttackers && moreAttackers.size() <= m)
				return groupSize2Freq;
			int step = 0;
			Set<Site> currentAttackers = null;
			
			// find a point where adding tuples from a source change the privacy
			while (isAnonymizedForLessAttackers != isAnonymizedForMoreAttackers 
					&& Math.abs(numMoreAttackers - numLessAttackers) > 1){

				step = Math.max(1, (numMoreAttackers - numLessAttackers) / 2);
				currentAttackers = cg.nextSuperSet(lessAttackers, step);
				
				// In case all (numLessAttackers+step)-combinations are checked
				while (currentAttackers == null && numLessAttackers + step < numMoreAttackers) {
					currentAttackers = cg.nextSuperSet(lessAttackers, ++step);
				}
				
				// if all combinations are checked
				if (numLessAttackers + step >= numMoreAttackers){
					break;
				}
				
				// report stats for the middle cardinality
				toReport = sum(currentAttackers);
				if (!attackers)
					toReport = allTuples - toReport;
				this.reportPrivacyCheck(groupSize2Freq, toReport);

				defenders = new HashSet<Site>(sites);
				defenders.removeAll(currentAttackers);
				defendedGroup = this.getTuples(defenders);
				boolean isAnonymized = this.checker.isAnonymized(defendedGroup);
				if (defenders.size() > 0 
						&& !isAnonymized && currentAttackers.size() <= m)
					return groupSize2Freq;

				// update lessAttackers or moreAttackers, as needed
				if (isAnonymized) {
					isAnonymizedForLessAttackers = isAnonymized;
					if (isAnonymizedForLessAttackers == isAnonymizedForMoreAttackers)
						break;
					lessAttackers = currentAttackers;
					numLessAttackers = currentAttackers.size();
				}
				else {
					isAnonymizedForMoreAttackers = isAnonymized;
					if (isAnonymizedForLessAttackers == isAnonymizedForMoreAttackers)
						break;
					moreAttackers = currentAttackers;
					numMoreAttackers = currentAttackers.size();
					
					// it won't be private, since for <= m attackers remaining
					// data is not private
					if (numMoreAttackers <= m)
						return groupSize2Freq;
				}
				
			}

			// Update combination generators - by forbidding groups of super
			// and sub combinations.
			cg.forbidThisAndAllSub(lessAttackers);
			mCombinationsChecker.forbidThisAndAllSub(lessAttackers);
			cg.forbidThisAndAllSuper(moreAttackers);
			mCombinationsChecker.forbidThisAndAllSuper(moreAttackers);

		}
		return groupSize2Freq;
	}

	// INC + DEC + MCOMB
	public HashMap<Integer, Integer> simulateBasedOnNumAttackers(int m,
			boolean attackers, boolean inc, boolean dec, boolean mcomb){

		int numSites = this.site2tuples.size();//this.actualNumSites;
		int[] possibleAttackers = new int[numSites + 3];
		int i = 0;
		for (int cardinality = 1; cardinality < numSites; cardinality++){
			if (inc && cardinality <= m)
				possibleAttackers[i++] = cardinality;
			if (dec && numSites - cardinality >= m)
				possibleAttackers[i++] = numSites - cardinality;
		}

		if (mcomb)
			possibleAttackers[i++] = m;
		return simulateChecking(m, possibleAttackers, false, attackers);
		
	}

//	@Deprecated
//	// Use simulateBasedOnNumAttackers
//	public HashMap<Integer, Integer> simulateIncreasingNumAttackers(int m, boolean attackers){
//		
//		int numSites = this.site2numTuples.size();//this.name2tuplesDistribution.size();
//		int[] possibleAttackers = new int[numSites-1];
//		int i = 0;
//		for (int cardinality = 1; cardinality <= m; cardinality++){
//			possibleAttackers[i++] = cardinality;
//		}
//		return simulateChecking(m, possibleAttackers, false, attackers);
//
//	}
//
//	@Deprecated
	// Use simulateBasedOnNumAttackers
//	public HashMap<Integer, Integer> simulateDecreasingNumAttackers(int m, boolean attackers){
//		int numSites = this.site2numTuples.size();//this.name2tuplesDistribution.size();
//		int[] possibleAttackers = new int[numSites];
//		int i = 0;
//		for (int cardinality = numSites; cardinality > 0; cardinality--){
//			possibleAttackers[i++] = cardinality;
//		}
//		return simulateChecking(m, possibleAttackers, false, attackers);
//	}

//	public static ArrayList<Integer> getEvenlyDistributedNumbers(int size, int sumAllValues){
//		ArrayList<Integer> numbers = new ArrayList<Integer>(size);
//		
//		for (int i = 0; i < size; i++){
//			numbers.add(sumAllValues / size);
//		}
//		
//		int reminder = size - sumAllValues % size;
//		for (int i = reminder; i < size; i++) {
//			numbers.set(i, numbers.get(i) + 1);
//		}
//
//		return numbers;
//	}
//
//	public static ArrayList<Integer> getExponentialyDistributedNumbers(int sumAllValues){
//		int size = (int) Math.log(sumAllValues + 1.0);
//		ArrayList<Integer> numbers = new ArrayList<Integer>(size);
//
//		for (int i = 0; i < size; i++){
//			numbers.add((int) Math.round(Math.exp(i)));
//		}
//		
//		int remaining = sumAllValues;
//		for (int i = 0; i < size; i++) {
//			remaining -= numbers.get(i);
//		}
//		
//		ArrayList<Integer> remainings = getEvenlyDistributedNumbers(size, remaining);
//		int sum = 0;
//		for (int i = 0; i < size; i++){
//			numbers.set(i, numbers.get(i) + remainings.get(i));
//			sum += remainings.get(i);
//		}
//
//		return numbers;
//	}
	
//	public int sum(Collection<String> collection) {
//		int sum = 0;
//		for (String name : collection){
//			if (this.name2tuplesDistribution.containsKey(name))
//				sum += this.name2tuplesDistribution.get(name);
//		}
//		return sum;
//	}
	
	public int sum(Collection<Site> collection) {
		int sum = 0;
		for (Site site : collection){
			Integer numTuples = this.site2numTuples.get(site);
			if (numTuples != null)
				sum += numTuples;
		}
		return sum;
	}
	
	private static void printStats(final String name, final HashMap<Integer, Integer> numTuples2freq) {
		Set<Integer> keys = numTuples2freq.keySet();
		List<Integer> sortedKeys = new ArrayList<Integer>(keys);
		Collections.sort(sortedKeys);
		for (int numTuples : sortedKeys){
			int freq = numTuples2freq.get(numTuples);
			System.out.println(name + "," + numTuples + "," + freq);
		}
		
	}

	private static HashMap<Integer, Integer> makeStatsGeneral(final HashMap<Integer, Integer> numTuples2freq, int rangeLength) {
		Set<Integer> keys = numTuples2freq.keySet();
		HashMap<Integer, Integer> range2freq = new HashMap<Integer, Integer>();
		for (Integer key : keys) {
			
			int generalKey = rangeLength * (int) Math.ceil(key / (double)rangeLength);
			if (!range2freq.containsKey(generalKey)){
				range2freq.put(generalKey, 0);
			}
			range2freq.put(generalKey, range2freq.get(generalKey) + numTuples2freq.get(key));
		}

		return range2freq;
	}
	
	private static long sumWithWeights(HashMap<Integer, Integer> numTuples2freq) {
		long sum = 0;
		
		if (numTuples2freq == null)
			return sum;
		
		for (Integer key : numTuples2freq.keySet())
			sum += (key * numTuples2freq.get(key)); 
		
		return sum;
	}

	private static long sumNumbers(Collection<Integer> s) {
		long sum = 0;
		
		for (Integer v : s)
			sum += v; 
		
		return sum;
	}
	
	private static void simulate(int m, int step, String name, CombinationsCounter cc) {
		
		Set<Integer> keys = new HashSet<Integer>();
		
		HashMap<Integer, Integer> numTuples2freqINC = cc.simulateBasedOnNumAttackers(m, true, true, false, false);//cc.simulateIncreasingNumAttackers(m, true);
		numTuples2freqINC = makeStatsGeneral(numTuples2freqINC, step);
		keys.addAll(numTuples2freqINC.keySet());
//		CombinationsCounter.printGeneralStats("INC", numTuples2freqINC, step);
		
		HashMap<Integer, Integer> numTuples2freqDEC = cc.simulateBasedOnNumAttackers(m, true, false, true, false);//cc.simulateDecreasingNumAttackers(m, true);
		numTuples2freqDEC = makeStatsGeneral(numTuples2freqDEC, step);
		keys.addAll(numTuples2freqDEC.keySet());
//		CombinationsCounter.printGeneralStats("DEC", numTuples2freqDEC, step);
				
		HashMap<Integer, Integer> numTuples2freqNSWITCH = cc.simulateCheckingWithoutSwitch(m, true);
		numTuples2freqNSWITCH = makeStatsGeneral(numTuples2freqNSWITCH, step);
		keys.addAll(numTuples2freqNSWITCH.keySet());
//		CombinationsCounter.printGeneralStats("NSWITCH", numTuples2freqNSWITCH, step);

		HashMap<Integer, Integer> numTuples2freqSWITCH = cc.simulateCheckingWithSwitch(m, true);
		numTuples2freqSWITCH = makeStatsGeneral(numTuples2freqSWITCH, step);
		keys.addAll(numTuples2freqSWITCH.keySet());
//		CombinationsCounter.printGeneralStats("SWITCH", numTuples2freqSWITCH, step);
		
		HashMap<Integer, Integer> numTuples2freqBIN = cc.simulateCheckingWithBinaryJumping(m, true);
		numTuples2freqBIN = makeStatsGeneral(numTuples2freqBIN, step);
		keys.addAll(numTuples2freqBIN.keySet());
//		CombinationsCounter.printGeneralStats("BIN", numTuples2freqBIN, step);
		
		HashMap<Integer, Integer> numTuples2freqMCOMB = cc.simulateBasedOnNumAttackers(m, true, false, false, true);
		numTuples2freqMCOMB = makeStatsGeneral(numTuples2freqMCOMB, step);
		keys.addAll(numTuples2freqMCOMB.keySet());
//		CombinationsCounter.printGeneralStats("MCOMB", numTuples2freqMCOMB, step);
		
		HashMap<Integer, Integer> numTuples2freqINCDECMCOMB = cc.simulateBasedOnNumAttackers(m, true, true, true, true);
		numTuples2freqINCDECMCOMB = makeStatsGeneral(numTuples2freqINCDECMCOMB, step);
		keys.addAll(numTuples2freqINCDECMCOMB.keySet());
//		CombinationsCounter.printGeneralStats("INC+DEC+MCOMB", numTuples2freqINCDECMCOMB, step);
		
		// getEvenlyDistributedNumbers(10, 100);

		List<Integer> sortedKeys = new ArrayList<Integer>(keys);
		Collections.sort(sortedKeys);
		for (Integer key : sortedKeys) {
			System.out.print(name + "," + m + "," + key);
			if (numTuples2freqINC.containsKey(key))
				System.out.print("," + numTuples2freqINC.get(key));
			else
				System.out.print(",");
			if (numTuples2freqDEC.containsKey(key))
				System.out.print("," + numTuples2freqDEC.get(key));
			else
				System.out.print(",");
			
			if (numTuples2freqNSWITCH.containsKey(key))
				System.out.print("," + numTuples2freqNSWITCH.get(key));
			else
				System.out.print(",");
			if (numTuples2freqSWITCH.containsKey(key))
				System.out.print("," + numTuples2freqSWITCH.get(key));
			else
				System.out.print(",");
			
			if (numTuples2freqBIN.containsKey(key))
				System.out.print("," + numTuples2freqBIN.get(key));
			else
				System.out.print(",");
			
			if (numTuples2freqMCOMB.containsKey(key))
				System.out.print("," + numTuples2freqMCOMB.get(key));
			else
				System.out.print(",");

			if (numTuples2freqINCDECMCOMB.containsKey(key))
				System.out.print("," + numTuples2freqINCDECMCOMB.get(key));
			else
				System.out.print(",");

			System.out.println();
		}

	}

	private static void simulateSumWOAdaptiveOrdering(int m, String name, CombinationsCounter cc) {
		long start = 0L;
		long end = 0L;
		
		cc.setAdaptiveOrdering(true);
		start = Calendar.getInstance().getTimeInMillis();
		HashMap<Integer, Integer> numTuples2freq = cc.simulateBasedOnNumAttackers(m, false, false, true, false);
		end = Calendar.getInstance().getTimeInMillis();
		long sumDEC = sumWithWeights(numTuples2freq);
		long numGroupsDEC = sumNumbers(numTuples2freq.values());
		double avgGroupSizeDEC = sumDEC / (double) numGroupsDEC;
		long timeDEC = end - start;

		cc.setAdaptiveOrdering(false);
		start = Calendar.getInstance().getTimeInMillis();
		HashMap<Integer, Integer> numTuplesWOAO2freq = cc.simulateBasedOnNumAttackers(m, false, false, true, false);
		end = Calendar.getInstance().getTimeInMillis();
		long sumWOAO = sumWithWeights(numTuplesWOAO2freq);
		long numGroupsWOAO = sumNumbers(numTuplesWOAO2freq.values());
		double avgGroupSizeWOAO = sumWOAO / (double) numGroupsWOAO;
		long timeWOAO = end - start;

		cc.setAdaptiveOrdering(true);
		start = Calendar.getInstance().getTimeInMillis();
		HashMap<Integer, Integer> numTuplesBIN2freq = cc.simulateCheckingWithBinaryJumping(m, false);
		end = Calendar.getInstance().getTimeInMillis();
		long sumBIN = sumWithWeights(numTuplesBIN2freq);
		long numGroupsBIN = sumNumbers(numTuplesBIN2freq.values());
		double avgGroupSizeBIN = sumBIN / (double) numGroupsBIN;
		long timeBIN = end - start;

		cc.setAdaptiveOrdering(false);
		start = Calendar.getInstance().getTimeInMillis();
		HashMap<Integer, Integer> numTuplesWOAOBIN2freq = cc.simulateCheckingWithBinaryJumping(m, false);
		end = Calendar.getInstance().getTimeInMillis();
		long sumWOAOBIN = sumWithWeights(numTuplesWOAOBIN2freq);
		long numGroupsWOAOBIN = sumNumbers(numTuplesWOAOBIN2freq.values());
		double avgGroupSizeWOAOBIN = sumWOAOBIN / (double) numGroupsWOAOBIN;
		long timeWOAOBIN = end - start;

		cc.setAdaptiveOrdering(true);
		
		System.out.println(name + "," + m + "," + cc.tuples.getNumUncorruptedTuples() 
				+ "," + timeWOAO+ "," + timeDEC + "," + timeWOAOBIN + "," + timeBIN
				+ "," + sumWOAO + "," + sumDEC + "," + sumWOAOBIN + "," + sumBIN
				+ "," + avgGroupSizeWOAO + "," + avgGroupSizeDEC+ "," + avgGroupSizeWOAOBIN + "," + avgGroupSizeBIN
				+ "," + numGroupsWOAO + "," + numGroupsDEC + "," + numGroupsWOAOBIN + "," + numGroupsBIN);
	}
	
	private static void simulateSum(int m, String name, CombinationsCounter cc) {
		long start = 0L;
		long end = 0L;
		
		start = Calendar.getInstance().getTimeInMillis();
		HashMap<Integer, Integer> numTuples2freq = cc.simulateBasedOnNumAttackers(m, false, true, false, false);
		end = Calendar.getInstance().getTimeInMillis();
		long sumINC = sumWithWeights(numTuples2freq);
		long numGroupsINC = sumNumbers(numTuples2freq.values());
		double avgGroupSizeINC = sumINC / (double) numGroupsINC;
		long timeINC = end - start;
		
		start = Calendar.getInstance().getTimeInMillis();
		numTuples2freq = cc.simulateBasedOnNumAttackers(m, false, false, true, false);
		end = Calendar.getInstance().getTimeInMillis();
		long sumDEC = sumWithWeights(numTuples2freq);
		long numGroupsDEC = sumNumbers(numTuples2freq.values());
		double avgGroupSizeDEC = sumDEC / (double) numGroupsDEC;
		long timeDEC = end - start;

//		start = Calendar.getInstance().getTimeInMillis();
//		numTuples2freq = cc.simulateCheckingWithoutSwitch(m, false);
//		end = Calendar.getInstance().getTimeInMillis();
//		long sumNSWITCH = sumWithWeights(numTuples2freq);
//		long numGroupsNSWITCH = sumNumbers(numTuples2freq.values());
//		double avgGroupSizeNSWITCH = sumNSWITCH / (double) numGroupsNSWITCH;
//		long timeNSWITCH = end - start;
//		
//		start = Calendar.getInstance().getTimeInMillis();
//		numTuples2freq = cc.simulateCheckingWithSwitch(m, false);
//		end = Calendar.getInstance().getTimeInMillis();
//		long sumSWITCH = sumWithWeights(numTuples2freq);
//		long numGroupsSWITCH = sumNumbers(numTuples2freq.values());
//		double avgGroupSizeSWITCH = sumSWITCH / (double) numGroupsSWITCH;
//		long timeSWITCH = end - start;

		start = Calendar.getInstance().getTimeInMillis();
		numTuples2freq = cc.simulateCheckingWithBinaryJumping(m, false);
		end = Calendar.getInstance().getTimeInMillis();
		long sumBIN = sumWithWeights(numTuples2freq);
		long numGroupsBIN = sumNumbers(numTuples2freq.values());
		double avgGroupSizeBIN = sumBIN / (double) numGroupsBIN;
		long timeBIN = end - start;

//		start = Calendar.getInstance().getTimeInMillis();
//		numTuples2freq = cc.simulateBasedOnNumAttackers(m, false, true, true, true);
//		end = Calendar.getInstance().getTimeInMillis();
//		long sumINCDECMCOMB = sumWithWeights(numTuples2freq);
//		long numGroupsINCDECMCOMB = sumNumbers(numTuples2freq.values());
//		double avgGroupSizeINCDECMCOMB = sumINCDECMCOMB / (double) numGroupsINCDECMCOMB;
//		long timeINCDECMCOMB = end - start;
		
		start = Calendar.getInstance().getTimeInMillis();
		numTuples2freq = cc.simulateBasedOnNumAttackers(m, false, false, false, true);
		end = Calendar.getInstance().getTimeInMillis();
		long sumMCOMB = sumWithWeights(numTuples2freq);
		long numGroupsMCOMB = sumNumbers(numTuples2freq.values());
		double avgGroupSizeMCOMB = sumMCOMB / (double) numGroupsMCOMB;
		long timeMCOMB = end - start;
		
		System.out.println(name + "," + m + "," + cc.tuples.getNumUncorruptedTuples() 
				+ "," + timeINC + "," + timeDEC + "," + + timeBIN + "," + timeMCOMB
				+ "," + sumINC + "," + sumDEC + "," + sumBIN + "," + sumMCOMB
				+ "," + avgGroupSizeINC + "," + avgGroupSizeDEC + "," + avgGroupSizeBIN + "," + avgGroupSizeMCOMB
				+ "," + numGroupsINC + "," + numGroupsDEC + "," + numGroupsBIN + "," + numGroupsMCOMB);
	}

	private static void predefinedDistributedRecords(double alpha, TuplesGroup group, K_anonymity k_anonymity,
			Entropy_L_diversity l_div, String name, boolean withHeader){
	
		PrivacyChecker checker = new AndCheckers.AndCheckersBuilder()
								.addChecker(k_anonymity, 1.0 - alpha)
								.addChecker(l_div,alpha)
								.build(name);
		CombinationsCounter cc = new CombinationsCounter(group, checker);

		cc.setAdaptiveOrdering(true);
		//System.out.println("SOURCE,FREQ,diversity,k-anonymity PRIVACY_LEVEL,entropy-l-diversity PRIVACY_LEVEL,PRIVACY_LEVEL");
		Map<Site, List<Tuple>> sites2tuples = group.getTuplesSplittedToSites(true);
//		Measure entropy = MeasuresFactory.getMeasure(Entropy.NAME);
		double avgPrivacyLevel = 0.0;
		List<Double> privacyLevels = new ArrayList<Double>();
		int actualNumSites = 0;
		for (Site site : sites2tuples.keySet()) {
			List<Tuple> tuples = sites2tuples.get(site);
			if (tuples.size() == 0)
				continue;
			double pl = cc.getPrivacyLevel(tuples);
			avgPrivacyLevel += pl;
			privacyLevels.add(pl);
			actualNumSites++;
//			System.out.println(site.getID() + "," + sites2tuples.get(site).size()
//					+ "," + Math.exp(entropy.getValue(new TuplesGroup(sites2tuples.get(site))))
//					+ "," + k_anonymity.getPrivacyFitness(new TuplesGroup(sites2tuples.get(site))) 
//					+ "," + l_div.getPrivacyFitness(new TuplesGroup(sites2tuples.get(site)))
////					+ "," + l_div.emed.getValue(new TuplesGroup(sites2tuples.get(site)))
//					+ "," + cc.getPrivacyLevel(sites2tuples.get(site)));
		}
		avgPrivacyLevel /= (double) actualNumSites;
		Collections.sort(privacyLevels);
		double median = (actualNumSites % 2 == 0 ? (privacyLevels.get(actualNumSites/2) + privacyLevels.get(actualNumSites/2-1) )/2.0 : privacyLevels.get(actualNumSites/2));
		double overallPrivacyLevel = cc.getPrivacyLevel(cc.getTuples(sites2tuples.keySet()).getTuplesAsList());
//		System.out.println("Avg privacy level: " + avgPrivacyLevel);
		name = "" + avgPrivacyLevel + "," + median + "," + overallPrivacyLevel + "," + name;

		if (withHeader){
			System.out.println("AVG PRIVACY FITNESS,MED PF,OVERALL PF,ALPHA,K,NUM SITES,L,M,NUM TUPLES,"
					 + "TIME,TIME,TIME,TIME," // time
					 + "SUM,SUM,SUM,SUM," //sum
					 + "AVG GROUP SIZE,AVG GROUP SIZE,AVG GROUP SIZE,AVG GROUP SIZE," // avg group size
					 + "NUM GROUPS,NUM GROUPS,NUM GROUPS,NUM GROUPS"); // num groups
			if (!cc.withAdaptiveOrdering)
				System.out.println("AVG PRIVACY FITNESS,MED PF,OVERALL PF,ALPHA,K,NUM SITES,L,M,NUM TUPLES,"
								 + "top-down random ordering,top-down,binary random ordering,binary," // time
								 + "top-down random ordering,top-down,binary random ordering,binary," // sum
								 + "top-down random ordering,top-down,binary random ordering,binary," // avg group size
								 + "top-down random ordering,top-down,binary random ordering,binary"); // num groups
			else
				System.out.println("AVG PRIVACY FITNESS,MED PF,OVERALL PF,ALPHA,K,NUM SITES,L,M,NUM_TUPLES,bottom-up,top-down,binary,direct," // num tuples
							 + "bottom-up,top-down,binary,direct," // time
							 + "bottom-up,top-down,binary,direct," // avg group size
							 + "bottom-up,top-down,binary,direct"); // num groups
		}
			//		System.out.println(name); 

		final int m = 5;
//		for (int m = 1; m <= actualNumSites / 2; m++){
			// simulate(m, step, privacyLevel, cc);
			if (!cc.withAdaptiveOrdering) {
				simulateSumWOAdaptiveOrdering(Math.min(actualNumSites - 1, m), name, cc);
//				System.out.println("Wo" + Math.min(actualNumSites - 1, m));
				if (actualNumSites - m > m) {
//					System.out.println(Math.min(actualNumSites - 1, actualNumSites - m));
//					simulateSumWOAdaptiveOrdering(Math.min(actualNumSites - 1, actualNumSites - m), name, cc);
				}
			} else {
//				System.out.println(Math.min(actualNumSites - 1, m));
				simulateSum(Math.min(actualNumSites - 1, m), name, cc);
				if (actualNumSites - m > m) {
//					System.out.println(Math.min(actualNumSites - 1, actualNumSites - m));
//					simulateSum(Math.min(actualNumSites - 1, actualNumSites - m), name, cc);
				}
			}
//		}

	}
	
//	private static void evenlyDistributedRecords(String filename){
////		List<Integer> tuples = getEvenlyDistributedNumbers(11, 100000);
//		List<Integer> tuples = readTuplesDistribution(filename);
//		Collections.sort(tuples);
//	
//		int i = 0;
//		System.out.println("SOURCE,FREQ");
//		for (int freq : tuples)
//			System.out.println(i++ + "," + freq);
//		System.out.println("PRIVACY_LEVEL,M,INC,DEC,NSWITCH,SWITCH,BIN,MCOMB");
//
//		final int step = 1000000;
//		for (int privacyLevel = 1000; privacyLevel <= 100000; privacyLevel += 10000){
//			CombinationsCounter cc = new CombinationsCounter(tuples, privacyLevel);
//			for (int m = 1; m < tuples.size(); m++) {
//				//simulate(m, step, privacyLevel, cc);
//				simulateSum(m, privacyLevel, cc, null);
//			}
//		}
//	}

//	private static void exponentiallyDistributedRecords(String filename){
//		List<Integer> tuples = readTuplesDistribution(filename);
//		Collections.sort(tuples);
////		List<Integer> tuples = getExponentialyDistributedNumbers(100000);
////		tuples.clear();
////		tuples.addAll(Arrays.asList(5924, 5926, 5930, 5943, 5978, 6071, 6326, 7020, 8905, 14027, 27950));
//		List<Integer> invTuples = new ArrayList<Integer>(tuples);
//		Collections.reverse(invTuples);
//		
//		int i = 0;
//		int sumTuples = 0;
//		System.out.println("SOURCE,FREQ");
//		for (int freq : tuples){
//			System.out.println(i++ + "," + freq);
//			sumTuples += freq;
//		}
////		System.out.println("PRIVACY_LEVEL,M,INC,DEC,invDEC,NSWITCH,SWITCH,invSWITCH,BIN,invBIN,MCOMB");
//		System.out.println("PRIVACY_LEVEL,M,INC,DEC,NSWITCH,SWITCH,BIN,MCOMB");
//
//		final int step = 1000000;
//		for (int privacyLevel = 1000; privacyLevel <= sumTuples; privacyLevel += step) {
//			CombinationsCounter cc = new CombinationsCounter(tuples, privacyLevel);
////			CombinationsCounter invcc = new CombinationsCounter(invTuples, privacyLevel);
//			for (int m = 1; m < tuples.size(); m++) {
//				//simulate(m, step, privacyLevel, cc, invcc);
//				simulateSum(m, privacyLevel, cc, null);
//			}
//		}
//	}


	public static List<Integer> readTuplesDistribution(String fileName){
		Map<Integer, Integer> id2freq = new HashMap<Integer, Integer>();
		int maxID = Integer.MIN_VALUE;
		BufferedReader reader;
		try
		{
			reader = new BufferedReader(new FileReader(fileName));
			String line;
			while ((line = reader.readLine()) != null) {
				String[] attrs = line.replaceAll(" ", "").split(",");
				int site = Integer.parseInt(attrs[attrs.length - 1]);
				Integer currentFreq = id2freq.get(site);
				maxID = Math.max(site, maxID);
				if (currentFreq == null)
					currentFreq = 0;
				id2freq.put(site, currentFreq+1);
			}
			reader.close();
		}
		catch (FileNotFoundException e)
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		catch (NumberFormatException e)
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		catch (IOException e)
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		
		List<Integer> freqs = new ArrayList<Integer>(maxID+1);
		Integer freq = null;
		int sumFreqs = 0;
		for (int i = 0; i <= maxID; i++){
			freq = id2freq.get(i);
			if (freq == null)
				freq = 0;
			freqs.add(freq);
			sumFreqs += freq;
		}
		
		System.out.println("Num tuples: " + sumFreqs);
		return freqs;
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args)
	{
		String filename = args[0]; // filename with tuples
//		final int k = Integer.parseInt(args[1]); // k-anonymity
//		final int l = Integer.parseInt(args[2]); // entropy-l-diversity
//		final int totalNumSites = Integer.parseInt(args[3]);
		final int maxNumSites =  Integer.parseInt(args[1]);
		int numTuplesPerSite = Integer.parseInt(args[2]);
		final double alpha = Double.parseDouble(args[3]);
		
		MeasuresFactory.registerMeasure(Entropy.NAME, new Entropy());
		try
		{
			boolean withHeader = true;
//			for (int numSites = 3; numSites <= maxNumSites; numSites++) {
			final int maxNumTuplesPerSite = numTuplesPerSite;
			for (numTuplesPerSite = 1; numTuplesPerSite <= maxNumTuplesPerSite; numTuplesPerSite += 1){
			final int numSites = maxNumSites;
//			for (int n = 50; n < 200; n+=10) {
				final int sampleSize = numSites*numTuplesPerSite;
				filename = args[0] + "_n_"+numSites+"_sample_"+sampleSize+"_SGadult.csv";
				
				// if file does not exist - sample it and save it for next runs
				if (!new File(filename).exists()) {
					RawTuples stats = RawTuples.readTuples("SGadult.csv", false);
					if (sampleSize > 0)
						stats.sampleTuples(sampleSize, true);
					if (args[0].equals("BA"))
						stats.randomNodeDistributionUsingBarabasiAlbertModel(numSites, 3, 0);
					else if (args[0].equals("Uniform"))
						stats.spreadTuplesUniformly(numSites);
					else
						throw new IllegalArgumentException("Illegal type of the random distribution. Allowed types: Uniform, BA");
					RawTuples.saveTuples(stats.getTuples(), filename, true);
				}

				RawTuples stats = RawTuples.readTuples(filename, true);
			
			List<Tuple> tuples = new ArrayList<Tuple>();
			Set<Site> chosenSites = new HashSet<Site>(numSites);
			for (Tuple tuple: stats.getTuples()) {
				Site s = tuple.getSite();
				if (chosenSites.size() < numSites && !chosenSites.contains(s))
					chosenSites.add(s);
				if(chosenSites.contains(s))
					tuples.add(tuple);
			}
			
//			List<Site> sortedSites = stats.getSites();
//			while (sortedSites.size() < totalNumSites) {
//				Site site = new Site(sortedSites.get(sortedSites.size() - 1).getID() + 1);
//				sortedSites.add(site);
//			}
//			TuplesGroup group = new TuplesGroup(stats.getTuples(), new HashSet<Site>(sortedSites), null);
			TuplesGroup group = new TuplesGroup(tuples);
//			group.getAttributeRanges();
			final int l = 9;
//			for (int l = 1; l < 15; l++){
			final int k = 50;
//			for (int k = 10; k <= 300; k += 10) {
//			for (double alpha: Arrays.asList(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9)) {
				K_anonymity k_anonymity = new K_anonymity(k);
				Entropy_L_diversity entropy_l_diversity = new Entropy_L_diversity(l);
//				T_closeness t_closeness = new T_closeness(t, new DeltaGrowthGroup(DeltaGrowthGroup.P_PRIOR.ALL_DATA_BASED));
				String name = ""+alpha+ "," + k+","+numSites+","+l;
				predefinedDistributedRecords(alpha, group, k_anonymity, entropy_l_diversity, name, withHeader);
				withHeader = false;
//			}
//			}
			}
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}
		
	}

}
