package edu.emory.mathcs;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import sun.reflect.generics.reflectiveObjects.NotImplementedException;

/**
 * Generates combinations of given elements.
 * @author Slawek Goryczka
 *
 * @param <T> a type of elements in combinations
 */
public class CombinationsGenerator<T>
{
	// next combination
	private BitSet bitItems;
	// current combination
	private BitSet current;
	// list of items to be returned in the combination
	final private List<T> items;
	private List<Integer> itemsIndexes;
	// number of items in the current combination
	private int cardinality;
	// indicates if there is a next combination to be generated
	private boolean hasNext;

	/**
	 * This and all super combinations are forbidden
	 */
	private Set<BitSet> forbiddenSuperCombinations;
	/**
	 * This and all sub combinations are forbidden
	 */
	private Set<BitSet> forbiddenSubCombinations;
	private boolean allForbidden;
	
	public CombinationsGenerator (Set<T> elements) {
		this(new ArrayList<T>(elements));
	}
	
	public CombinationsGenerator (List<T> elements) {
		this.items = elements;
		this.itemsIndexes = new ArrayList<Integer>(this.items.size());
		for (int i = 0; i < this.items.size(); i++)
			this.itemsIndexes.add(i);
		this.bitItems = new BitSet(elements.size());
		this.cardinality = 0;
		this.hasNext = false;
		this.forbiddenSuperCombinations = new HashSet<BitSet>();
		this.forbiddenSubCombinations = new HashSet<BitSet>();
		this.current = null;
		this.allForbidden = false;
	}
	
	private synchronized void setNextWithPreferences(Set<T> nextCandidate, int change) {
		this.bitItems.clear();
		
		for (int i = 0; i < this.items.size(); i++) {
			if (nextCandidate.contains(this.items.get(i))) {
				this.bitItems.set(i);
			}
		}
		
		// add/remove bits
		if (change < 0) {
			int numChanges = 0;
			for (int i = this.itemsIndexes.size(); i >= 0; i--) {
				if (numChanges >= change)
					break;
				
				if (this.bitItems.get(this.itemsIndexes.get(i)) == false)
					continue;
				
				this.bitItems.set(this.itemsIndexes.get(i), false);
				numChanges++;
			}
		}
		else if (change > 0) {
			int numChanges = 0;
			for (int i = 0; i < this.itemsIndexes.size(); i++) {
				if (numChanges >= change)
					break;
				
				if (this.bitItems.get(this.itemsIndexes.get(i)) == true)
					continue;
				
				this.bitItems.set(this.itemsIndexes.get(i), true);
				numChanges++;
			}
		}

		this.cardinality = this.bitItems.cardinality();
		this.hasNext = (this.cardinality > 0);

		// to set this.current and this.bitItems (to next combination)
		if (change == 0 || isForbidden(this.bitItems)) {
			this.nextCombination();
			
			// smart way doesn't get a good solution - do the usual one
			if (this.hasNext() == false) {

				this.bitItems.clear();
				for (int i = 0; i < this.items.size(); i++) {
					if (nextCandidate.contains(this.items.get(i))) {
						this.bitItems.set(i);
					}
				}
				
				// add/remove bits
				if (change < 0) {
					int numChanges = 0;
					for (int i = 0; i < this.getMaxCardinality() && numChanges < -change; i++) {
						if (this.bitItems.get(i) == true) {
							this.bitItems.set(i, false);
							numChanges++;
						}
					}
				}
				else if (change > 0) {
					int numChanges = 0;
					for (int i = 0; i < this.getMaxCardinality() && numChanges < change; i++) {
						if (this.bitItems.get(i) == false) {
							this.bitItems.set(i, true);
							numChanges++;
						}
					}
				}
				if (change == 0 || isForbidden(this.bitItems))
					this.nextCombination();
			}
		}
	}

	/**
	 * Set next combination for a given one by changing current cardinality.
	 * @param nextCandidate current combination
	 * @param change change in the cardinality
	 */
	private synchronized void setNext(final Set<T> nextCandidate, int change)
	{

		this.bitItems = convertToBitSet(nextCandidate);

		// add/remove bits
		if (change < 0)
		{
			int numChanges = 0;
			for (int i = 0; i < this.getMaxCardinality() && numChanges < -change; i++)
			{
				if (this.bitItems.get(i) == true)
				{
					this.bitItems.set(i, false);
					numChanges++;
				}
			}
		}
		else if (change > 0)
		{
			int numChanges = 0;
			for (int i = 0; i < this.getMaxCardinality() && numChanges < change; i++)
			{
				if (this.bitItems.get(i) == false)
				{
					this.bitItems.set(i, true);
					numChanges++;
				}
			}
		}
		this.cardinality = this.bitItems.cardinality();
		this.hasNext = (this.cardinality > 0);
		
		if (change == 0 || isForbidden(this.bitItems))
			this.nextCombination();
	}

	/**
	 * Get weight for a given candidate. If it's in current, then it will be
	 * subtracted, otherwise added.
	 * @param current
	 * @param candidate
	 * @return percentage of not forbidden (to be checked) supersets.
	 */
	private synchronized double getWeight(final Set<T> current, final T candidate) {
		double weight = 0.0;
		BigInteger numCombinations = BigInteger.ZERO;
		Set<T> currentWithCandidate = new HashSet<T>(current);
		currentWithCandidate.add(candidate);
		BitSet currentBitSet = this.convertToBitSet(currentWithCandidate);
		BigInteger overallNumCombinations = this.getNumSubsets(currentBitSet);
		
		boolean toAdd = true;
		
		List<BitSet> forbiddenSubsets = this.getForbiddenSubsets(currentBitSet);
		
		while (forbiddenSubsets.size() > 1) {
			for (BitSet forbidden : forbiddenSubsets) {
				BigInteger numSupersets = this.getNumSupersets(forbidden);
				System.out.print((toAdd ? "+" : "-") + numSupersets);
				if (toAdd)
					numCombinations = numCombinations.add(numSupersets);
				else
					numCombinations = numCombinations.subtract(numSupersets);
			}
			forbiddenSubsets = this.getOverlappingSupersets(forbiddenSubsets);
			toAdd = !toAdd;
		}
		System.out.println();
		
		double overall = overallNumCombinations.doubleValue();
		if (overall == 0)
			return 0.0;
		double num = numCombinations.doubleValue();
		weight = num / overall;
		return 1.0 - weight;
	}
	
	/**
	 * For each pair of sets find their common superset and collect them.
	 * @param currentBitSets
	 * @return
	 */
	private synchronized List<BitSet> getOverlappingSupersets(final List<BitSet> currentBitSets){
		List<BitSet> forbiddenSupersets = new ArrayList<BitSet>();

		if (currentBitSets.size() < 2)
			return forbiddenSupersets;
		
		CombinationsGenerator<BitSet> pairsGenerator = new CombinationsGenerator<BitSet>(currentBitSets);
		pairsGenerator.setCardinality(2);
		while (pairsGenerator.hasNext()){
			BitSet[] setsToOverlap = pairsGenerator.next().toArray(new BitSet[0]);
			if (setsToOverlap.length < 2){
				forbiddenSupersets.add(setsToOverlap[0]);
				break;
			}
			BitSet commonSet = (BitSet) setsToOverlap[0].clone();
			commonSet.or(setsToOverlap[1]);
			forbiddenSupersets.add(commonSet);
		}

		return forbiddenSupersets;
		
	}
	
	private synchronized List<BitSet> getForbiddenSupersets(final BitSet currentBitSet) {
		List<BitSet> forbiddenSupersets = new ArrayList<BitSet>();
		
		for (BitSet bs : this.forbiddenSuperCombinations){
			if (currentBitSet.cardinality() > bs.cardinality())
				continue;
			
			BitSet bs2 = (BitSet) bs.clone();
			bs2.and(currentBitSet);
			if (bs2.cardinality() == currentBitSet.cardinality())
			{
				forbiddenSupersets.add(bs);
			}
		}
		
		return forbiddenSupersets;
	}
	
	private synchronized List<BitSet> getForbiddenSubsets(final BitSet currentBitSet) {
		List<BitSet> forbiddenSubsets = new ArrayList<BitSet>();
		
		for (BitSet bs : this.forbiddenSubCombinations){
			if (currentBitSet.cardinality() > bs.cardinality())
				continue;
			
			BitSet bs2 = (BitSet) bs.clone();
			bs2.and(currentBitSet);
			if (bs2.cardinality() == currentBitSet.cardinality())
			{
				forbiddenSubsets.add(bs);
			}
		}
		
		return forbiddenSubsets;
	}
	
	private synchronized BigInteger getNumSupersets(final BitSet current) {
		BigInteger numSupersets = BigInteger.ONE;
		int maxNumItems = this.getMaxCardinality();
		int numItems = current.cardinality();
		for (long i = 1; i <= maxNumItems - numItems; i++)
			numSupersets = numSupersets.multiply(BigInteger.valueOf(i));
		
		return numSupersets;
	}
	
	private synchronized BigInteger getNumSubsets(final BitSet current) {
		BigInteger numSupersets = BigInteger.ONE;
		int numItems = current.cardinality();
		for (long i = numItems; i > 0; --i)
			numSupersets = numSupersets.multiply(BigInteger.valueOf(i));
		
		return numSupersets;
	}
	
	/**
	 * Use number of super/sub combinations for each bit to weight them.
	 * Pick ones with the max. weight.
	 * @param nextCandidate
	 * @param change
	 */
	private synchronized void setNextSmart(final Set<T> nextCandidate, int change)
	{

		this.bitItems = convertToBitSet(nextCandidate);

		// add/remove bits
		if (change < 0)
		{
			int numChanges = 0;
			for (int i = 0; i < this.getMaxCardinality() && numChanges < -change; i++)
			{
				if (this.bitItems.get(i) == true)
				{
					this.bitItems.set(i, false);
					numChanges++;
				}
			}
		}
		else if (change > 0)
		{
			int numChanges = 0;
			for (int i = 0; i < this.getMaxCardinality() && numChanges < change; i++)
			{
				if (this.bitItems.get(i) == false)
				{
					this.bitItems.set(i, true);
					numChanges++;
				}
			}
		}
		
		this.cardinality = this.bitItems.cardinality();
		this.hasNext = (this.cardinality > 0);
		
		if (change == 0 || isForbidden(this.bitItems))
			this.nextCombination();
	}

	
	/**
	 * Doubled functionality wih nextSet(Set<T>);
	 * Get next combination which intersect the most with give one.
	 * @param toIntersectWith
	 * @return
	 */
	@Deprecated
	public synchronized Set<T> nextMaxIntersecting(final Set<T> toIntersectWith) {
		this.bitItems = this.convertToBitSet(toIntersectWith);
		
		int itemsToRemove = this.bitItems.cardinality() - this.getCardinality();

		// remove bits if there are too many of them
		for (; itemsToRemove > 0; itemsToRemove--)
			this.bitItems.clear(this.bitItems.nextSetBit(0));
/*		for (int i = this.bitItems.cardinality(); i >= 0 && itemsToRemove > 0; i--) {
			if (this.bitItems.get(i)) {
				this.bitItems.clear(i);
				itemsToRemove--;
			}
		}
*/
		// set bits to get a correct cardinality
		for (; itemsToRemove < 0; itemsToRemove++)
			this.bitItems.set(this.bitItems.nextClearBit(0));

		this.hasNext = (this.cardinality > 0);
		
		if (isForbidden(this.bitItems))
			this.nextCombination();

		return this.next();
	}
	
	/**
	 * Get next combination for a given one.
	 * @param current current combination
	 * @return next (for the current) combination or null if it does not exist
	 */
	public synchronized Set<T> next(Set<T> current) {
		this.setNext(current, 0);
		return this.next();
	}

	/**
	 * Get a next superset combination that contains the current one. 
	 * @param current the current combination
	 * @param add an increase of the cardinality
	 * @return a next combination that is a superset of the given one and
	 * an increased cardinality or null if it does not exist
	 */
	public synchronized Set<T> nextSuperSet(final Set<T> current, int add) {
		this.setNext(current, add);
		return this.next();
	}

	/**
	 * Get a next subset combination that is contained by the current one.
	 * @param current the current combination
	 * @param remove a decrease of the cardinality  
	 * @return a next combination that is a subset of the given one and
	 * a decreased cardinality or null if it does not exist
	 */
	public synchronized Set<T> nextSubSet(final Set<T> current, int remove) {
		this.setNext(current, -remove);
		return this.next();
	}

	/**
	 * Get a next combination.
	 * @return a next combination of given items or null if it does not exist
	 */
	public synchronized Set<T> next() {
		
		if (this.hasNext() == false)
			return null;
		
		Set<T> combination = new HashSet<T>();
		
		for (int i = 0; combination.size() < this.cardinality && i < this.items.size(); i++)
		{
			if (this.bitItems.get(i))
				combination.add(this.items.get(i));
			
		}
		
		// set the next combination
		this.nextCombination();
		
		return combination;
	}
	
	/**
	 * Check if there is a next combination to be generated.
	 * @return true if there is a next combination to be generated, false otherwise
	 */
	public synchronized boolean hasNext() {
		return hasNext;
	}

	/**
	 * Set a new cardinality to the combination generator.
	 * @param cardinality new cardinality of combinations that are generated
	 */
	public synchronized void setCardinality(int cardinality) {
		this.cardinality = cardinality;
		this.hasNext = (cardinality > 0);
		this.bitItems.clear();

		this.bitItems.set(0, this.cardinality);
		if (this.isForbidden(this.bitItems))
			this.nextCombination();
	}
	
	/**
	 * Get a current cardinality of generated combinations.
	 * @return current cardinality of combinations
	 */
	public synchronized int getCardinality() {
		return this.cardinality;
	}
	
	/**
	 * Get max possible cardinality (size of the longest combination that may be
	 * generated) -- get number of elements used to generate combinations.
	 * @return max possible cardinality
	 */
	public synchronized int getMaxCardinality() {
		return this.items.size();
	}
	
	/**
	 * Forbid the current combination and all supercombinations.
	 */
	public synchronized void forbidCurrentAndAllSuper() {
		this.forbidThisAndAllSuper(this.current);
	}

	/**
	 * Forbid a given combination and all its super combinations
	 * @param bs
	 */
	private synchronized void forbidThisAndAllSuper(final BitSet bs) {
		// don't multiply forbidden sets
		if (this.isForbiddenSuper(bs))
			return;
		
		// Make sure that the set of forbidden combinations is minimal,
		// by removing ones that repeats
		Set<BitSet> toBeRemoved = new HashSet<BitSet>();
		for (BitSet b : this.forbiddenSuperCombinations) {
			if (b.cardinality() <= bs.cardinality())
				continue;
			
			BitSet b2 = (BitSet) b.clone();
			b2.and(bs);
			if (b2.cardinality() == bs.cardinality())
				toBeRemoved.add(b);
		}
		
		this.forbiddenSuperCombinations.removeAll(toBeRemoved);
		this.forbiddenSuperCombinations.add(bs);
	}

	/**
	 * Forbid current and all its subcombinations.
	 */
	public synchronized void forbidCurrentAndAllSub() {
		this.forbidThisAndAllSub(this.current);
	}

	/**
	 * Forbid a given and all its subcombinations
	 * @param bs a combination that should be forbidden togheter with its
	 * subcombinations
	 */
	private synchronized void forbidThisAndAllSub(final BitSet bs) {
		// don't multiply forbidden sets
		if (this.isForbiddenSub(bs)) {
			return;
		}
		
		// Make sure that the set of forbidden subcombinations is minimal,
		// by removing redundant combinations.
		Set<BitSet> toBeRemoved = new HashSet<BitSet>();
		for (BitSet b : this.forbiddenSubCombinations) {
			if (b.cardinality() > bs.cardinality())
				continue;
			
			BitSet b2 = (BitSet) b.clone();
			b2.and(bs);
			if (b2.cardinality() == b.cardinality())
				toBeRemoved.add(b);
		}
		
		this.forbiddenSubCombinations.removeAll(toBeRemoved);
		this.forbiddenSubCombinations.add(bs);
	} 
	
	/**
	 * 
	 * @param bs
	 * @return
	 */
	private synchronized boolean isForbidden(final BitSet bs) {
		return isForbiddenSuper(bs) || isForbiddenSub(bs);
	}

	private synchronized boolean isForbiddenSub(final BitSet bs) {
		for (BitSet forbidden : this.forbiddenSubCombinations) {
			if (forbidden.cardinality() < bs.cardinality())
				continue;
			
			BitSet forbidden2 = (BitSet) forbidden.clone();
			forbidden2.and(bs);
			if (forbidden2.cardinality() == bs.cardinality())
			{
				return true;
			}
		}
		
		return false;
	}
	
	// Check if bs is forbidden as a superset
	private synchronized boolean isForbiddenSuper(final BitSet bs) {
		for (BitSet forbidden : this.forbiddenSuperCombinations) {
			if (forbidden.cardinality() > bs.cardinality())
				continue;
			
			BitSet bs2 = (BitSet) bs.clone();
			bs2.and(forbidden);
			if (bs2.cardinality() == forbidden.cardinality())
			{
				return true;
			}
		}
		
		return false;
	}
	
	public synchronized int getNumForbiddenCombinations() {
		return this.forbiddenSubCombinations.size() + this.forbiddenSuperCombinations.size();
	}
	
	private synchronized void nextCombination() {
		this.current = (BitSet) this.bitItems.clone();
		
		do {
			
			int firstSetBit = this.bitItems.nextSetBit(0);
			int firstClearBit = this.bitItems.nextClearBit(firstSetBit);

			if (firstClearBit >= this.getMaxCardinality())
				this.hasNext = false;
			
			this.bitItems.set(firstClearBit);
			this.bitItems.clear(0, firstClearBit);
	
			int currentCardinality = this.bitItems.cardinality();
			this.bitItems.set(0, this.cardinality - currentCardinality);
		}
		while (this.hasNext && this.isForbidden(this.bitItems));
	}

	@Deprecated
	public synchronized boolean canMcombinationsBeDecided(int m) {
		BitSet superSet = new BitSet(this.getMaxCardinality());
		for (BitSet superSets : this.forbiddenSuperCombinations) {
			if (superSets.cardinality() > m)
				continue;
			
			superSet.or(superSets);
		}
		
		if (superSet.cardinality() == this.getMaxCardinality())
			return true;
	
		for (BitSet subSets : this.forbiddenSubCombinations) {
			if (subSets.cardinality() >= m)
				return true;
		}
		
		return false;
	}
	
	public synchronized boolean isThisOrAnySuperCombinationForbidden(int k) {
		for (BitSet subSets : this.forbiddenSubCombinations) {
			if (subSets.cardinality() >= k)
				return true;
		}
		
		return false;
	}

	private synchronized BitSet convertToBitSet(Set<T> prevSites) {
		
		BitSet bs = new BitSet(this.getMaxCardinality());
		for (T s : prevSites) {
			int i = this.items.indexOf(s);
			if (i >= 0) {
				bs.set(i);
			}
		}
		return bs;
	}
	
	public synchronized void forbidThisAndAllSub(Set<T> sites)
	{
		if (sites == null)
			return;
		
		this.forbidThisAndAllSub(this.convertToBitSet(sites));

		// check if you didn't forbid next one by any chance
		if (this.hasNext() && this.isForbidden(this.bitItems))
			this.nextCombination();
	}

	public synchronized void forbidThisAndAllSuper(Set<T> sites)
	{
		if (sites == null)
			return;
		
		this.forbidThisAndAllSuper(this.convertToBitSet(sites));
		
		// check if you didn't forbid next one by any chance
		if (this.hasNext() && this.isForbidden(this.bitItems))
			this.nextCombination();
	}

	public synchronized Set<T> all()
	{
		return new HashSet<T>(this.items);
	}

	public synchronized boolean areThisAndAllNextDecided()
	{
		if (!this.hasNext())
			return true;
		
		// set bitItems to the next unforbidden set
//		this.nextCombination();
//
//		if (!this.hasNext)
//			return true;

		return false;
	}

	public synchronized Set<T> peek()
	{
		Set<T> list = new HashSet<T>(this.bitItems.cardinality());
		
		for (int i = 0; i < this.bitItems.cardinality(); i++)
			if (this.bitItems.get(i))
				list.add(this.items.get(i));
		
		return list;
	}

	public synchronized Set<T> nextSet(Set<T> nextCandidate, int change, Set<T> approxTarget)
	{
		this.bitItems.clear();
		for (int i = 0; i < this.items.size(); i++) {
			if (nextCandidate.contains(this.items.get(i))) {
				this.bitItems.set(i);
			}
		}
		
		BitSet approx = convertToBitSet(approxTarget);
		
		// add/remove bits
		if (change < 0) {
			int numChanges = 0;
			BitSet possibleChange = new BitSet(this.getMaxCardinality());
			// set true if bit is used there
			possibleChange.or(this.bitItems);
			
			BitSet preferedChange = approx;
			preferedChange.and(possibleChange);
			// add preferred changes
			for (int i = 0; i < this.getMaxCardinality() && numChanges < -change; i++) {
				if (preferedChange.get(i)) {
					this.bitItems.set(i, false);
					possibleChange.set(i, false);
					numChanges++;
				}
			}
			
			// add missing possible changes
			for (int i = 0; i < this.getMaxCardinality() && numChanges < -change; i++) {
				if (possibleChange.get(i)) {
					this.bitItems.set(i, false);
					possibleChange.set(i, false);
					numChanges++;
				}
			}
		}
		else if (change > 0) {
			int numChanges = 0;
			BitSet possibleChange = new BitSet(this.getMaxCardinality());
			// set true if bit is not used there
			possibleChange.xor(this.bitItems);
			possibleChange.flip(0, this.getMaxCardinality());
			
			BitSet preferedChange = approx;
			preferedChange.and(possibleChange);
			// add preferred changes
			for (int i = 0; i < this.getMaxCardinality() && numChanges < change; i++) {
				if (preferedChange.get(i)) {
					this.bitItems.set(i, true);
					possibleChange.set(i, false);
					numChanges++;
				}
			}
			
			// add missing possible changes
			for (int i = 0; i < this.getMaxCardinality() && numChanges < change; i++) {
				if (possibleChange.get(i)) {
					this.bitItems.set(i, true);
					possibleChange.set(i, false);
					numChanges++;
				}
			}
			
		}

		this.cardinality += change;
		
		if (this.isForbidden(this.bitItems))
			this.nextCombination();
		
		if (!this.hasNext())
			return null;

		Set<T> combination = new HashSet<T>();
		
		for (int i = 0; combination.size() < this.cardinality && i < this.items.size(); i++)
		{
			if (this.bitItems.get(i))
				combination.add(this.items.get(i));
		}
		
		return combination;
	}

	public synchronized Set<T> nextFlipped()
	{
		
		if (this.hasNext() == false)
			return null;
		
		Set<T> combination = new HashSet<T>();
		
		for (int i = 0; combination.size() < this.cardinality && i < this.items.size(); i++)
		{
			if (this.bitItems.get(i))
				combination.add(this.items.get(i));
			
		}
		
		BitSet original = this.bitItems;
		int numChanges = 0;
		BitSet flipped = new BitSet(this.getMaxCardinality());
		for (int i = this.getMaxCardinality(); i >= 0  && numChanges < this.cardinality; i--)
			if (this.bitItems.get(i) == false)
			{
				flipped.set(i, true);
				numChanges++;
			}
		this.bitItems = flipped;
		
		// set the next combination
		this.nextCombination();
		
		if (this.hasNext() == false)
		{
			this.bitItems = original;
			this.nextCombination();
		}
		
		return combination;
	}

	public synchronized boolean areAllAllCombinationsForbidden() {
		return this.allForbidden;
	}

	public synchronized void forbidAllCombinations()
	{
		this.allForbidden = true;
	}

	public synchronized Set<Set<Site>> nextCombinations(int numThreads)
	{
		throw new NotImplementedException();
	}

}
