package edu.emory.mathcs;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.Calendar;

import edu.emory.mathcs.anonymizers.Anonymizer;
import edu.emory.mathcs.anonymizers.ParallelAdaptiveMPrivacySplitterThreads;
import edu.emory.mathcs.measures.Count_diverse;
import edu.emory.mathcs.measures.Measure;
import edu.emory.mathcs.measures.MeasuresFactory;
import edu.emory.mathcs.measures.MinCountSites;
import edu.emory.mathcs.measures.privacy.DeltaGrowthAverage;
import edu.emory.mathcs.measures.privacy.DeltaGrowthGroup;
import edu.emory.mathcs.measures.privacy.DeltaGrowthValue;
import edu.emory.mathcs.measures.privacy.DisclosureProbabilityAvg;
import edu.emory.mathcs.measures.privacy.DisclosureProbabilityMax;
import edu.emory.mathcs.measures.utility.Entropy;
import edu.emory.mathcs.measures.utility.Entropy_diverse;
import edu.emory.mathcs.measures.utility.LossMetric;
import edu.emory.mathcs.measures.utility.ModifiedEntropy;
import edu.emory.mathcs.privacy.AndCheckers;
import edu.emory.mathcs.privacy.K_anonymity;
import edu.emory.mathcs.privacy.L_diversity;
import edu.emory.mathcs.privacy.M_privacy;
import edu.emory.mathcs.privacy.PrivacyChecker;

public class Main {
	public static void main(String[] args) throws IOException, InterruptedException {
		
		if (args.length != 10) {
			System.err.println("Usage:");
			System.err.println("  java Main <input data filename> <output data filename> "
					+ "<number of providers> <m for m-privacy> <k for k-anonymity> "
					+ "<l for l-diversity> <weight of l-diversity in the fitness score> "
					+ "<fitness threshold for activating top-down and de-activating binary verififers> "
					+ "<number of threads for verifiers> <number of threads for anonymizers>");
			System.exit(-1);
		}
		
		MeasuresFactory.registerMeasure(Entropy.NAME, new Entropy());
		MeasuresFactory.registerMeasure(ModifiedEntropy.NAME, new ModifiedEntropy());
		MeasuresFactory.registerMeasure(DeltaGrowthValue.NAME, new DeltaGrowthValue(DeltaGrowthValue.P_PRIOR.ZERO));
		MeasuresFactory.registerMeasure(LossMetric.NAME, new LossMetric());
		MeasuresFactory.registerMeasure(DeltaGrowthGroup.NAME, new DeltaGrowthGroup(DeltaGrowthGroup.P_PRIOR.EQUAL));
		MeasuresFactory.registerMeasure(DeltaGrowthAverage.NAME, new DeltaGrowthAverage(DeltaGrowthAverage.P_PRIOR.EQUAL));
		MeasuresFactory.registerMeasure(DisclosureProbabilityAvg.NAME, new DisclosureProbabilityAvg());
		MeasuresFactory.registerMeasure(DisclosureProbabilityMax.NAME, new DisclosureProbabilityMax());
		MeasuresFactory.registerMeasure(Entropy_diverse.NAME, new Entropy_diverse());
		MeasuresFactory.registerMeasure(MinCountSites.NAME, new MinCountSites());
		MeasuresFactory.registerMeasure(Count_diverse.NAME, new Count_diverse());
		
		// Input params
		final String inputDataFileName = args[0];
		final String outputDataFileName = args[1];
		final int siteNum = Integer.parseInt(args[2]);
		final int m = Integer.parseInt(args[3]);
		final int k = Integer.parseInt(args[4]);
		final int l = Integer.parseInt(args[5]);
		final double alpha = Double.parseDouble(args[6]);
		final double privacyVerifiersThreshold = Double.parseDouble(args[7]);
		final int numVerifyingThreads = Integer.parseInt(args[8]);
		final int numSplittingThreads = Integer.parseInt(args[9]);

		double total_time_s = -Calendar.getInstance().getTime().getTime();

		// Setting up verifiers
		PrivacyChecker l_diversity_checker = new L_diversity(l);
		PrivacyChecker k_anonymity = new K_anonymity(k); 
		PrivacyChecker k_anonymity_l_diversity_checker = new AndCheckers.AndCheckersBuilder()
						.addChecker(k_anonymity, 1.0 - alpha)
						.addChecker(l_diversity_checker, alpha)
						.build(k + "_anonymity_entropy" + l + "_diversity");

		PrivacyChecker dm_checker = new M_privacy(m, k_anonymity_l_diversity_checker, privacyVerifiersThreshold, numVerifyingThreads);
		Anonymizer mAnonymizer = new ParallelAdaptiveMPrivacySplitterThreads(k, dm_checker, numSplittingThreads);
		
		// Reading data
		RawTuples stats = RawTuples.readTuples(inputDataFileName, true);
		stats.distributeTuples(Tuple.RANDOM_UNIFORM, siteNum);
		TuplesGroup group = new TuplesGroup(stats);
		
		// Anonymizing data
		mAnonymizer.splitGroup(group);
			
		total_time_s += Calendar.getInstance().getTime().getTime();
		total_time_s /= 1000.0;
		System.out.println("Done in " + total_time_s + " seconds.");

		// Saving data into the output file
		Writer output = new BufferedWriter(new FileWriter(outputDataFileName));
		group.printGroups(output, false);
		output.flush();
		output.close();
	}
	
	private static void printPrivacyLevelStats(TuplesGroup tuples, int k, double l, Measure l_div) {
		if (tuples.hasSubGroups())
		{
			printPrivacyLevelStats(tuples.getLeftChild(), k, l, l_div);
			printPrivacyLevelStats(tuples.getRightChild(), k, l, l_div);
		} else {
			if (tuples.getNumTuples() < k || l_div.getValue(tuples) < l)
				System.out.println(tuples.getNumTuples() + "," + l_div.getValue(tuples));
		}
	}
}
