package edu.emory.mathcs;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class RawTuples {

	private List<Tuple> tuples;
	private int numAttributes;
	private List<IntAttributeRange> attributeRanges;
	
	private RawTuples(List<Tuple> tuples)
	{
		this.tuples = tuples;
		this.attributeRanges = new ArrayList<IntAttributeRange>();
		
		this.computeStatistics();
	}

	public void addTupleIfNeeded(Tuple tuple) {
		if (this.tuples.contains(tuple))
			return;
			
		this.tuples.add(tuple);
		for (int i = 0; i < this.attributeRanges.size(); i++)
			this.attributeRanges.get(i).updateFor(tuple.getData(i));
	}
	
	private void computeStatistics() 
	{
		if (this.tuples.size() == 0)
			return;
		
		this.numAttributes = this.tuples.get(0).getData().length;
		for (int i = 0; i < this.numAttributes; i++)
			this.attributeRanges.add(new IntAttributeRange(Integer.MAX_VALUE, Integer.MIN_VALUE));
		
		for (Tuple t : this.tuples)
		{
			if (this.numAttributes != t.getData().length)
				throw new RuntimeException("A tuple has different number of attributes!\n"+t.toString());
			
			for (int attributeIndex = 0; attributeIndex < this.numAttributes; attributeIndex++)
			{
				IntAttributeRange currentRange = this.attributeRanges.get(attributeIndex);
				int currentValue = t.getData(attributeIndex);
				currentRange.setMinMax(Math.min(currentValue, currentRange.getMin()), 
						Math.max(currentValue, currentRange.getMax()));
			}
			
			
		}
	}
	
	public List<Site> getSites() {
		Set<Site> sitesInSet = new HashSet<Site>();
		for (Tuple t: this.tuples){
			sitesInSet.add(t.getSite());
		}
		
		List<Site> sites = new ArrayList<Site>(sitesInSet);
		Collections.sort(sites);
		return sites;
	}
	
	public int getNumAllAttributes()
	{
		return this.numAttributes;
	}
	
	public IntAttributeRange getAttributeDomain(int index)
	{
		return this.attributeRanges.get(index);
	}

	public List<Tuple> getTuples() {
		return this.tuples;
	}

	public IntAttributeRange[] getAttributeRanges() {
		return this.attributeRanges.toArray(new IntAttributeRange[0]);
	}
	
	public void distributeTuples(int randType, int numOwners)
	{
		for (Tuple t : this.tuples)
			t.setSite(randType, numOwners);
	}
	
	public void spreadTuplesUniformly(int numOwners) {
		int numTuples = this.tuples.size();
		//int[] frequencies = new int[numOwners];
		HashMap<Integer, Integer> site2Freqs = new HashMap<Integer, Integer>();
		for (int i = 0; i < numOwners; i++){
			site2Freqs.put(i, numTuples / numOwners + (i >= numTuples % numOwners ? 0 : 1));
		}
		
		for (Tuple t: this.tuples) {
			List<Integer> ids = new ArrayList<Integer>();
			ids.addAll(site2Freqs.keySet());
			int i = (int)(Math.random() * ids.size());
			int id = ids.get(i);
			Site site = SitesFactory.getSite(id);
			site2Freqs.put(id, site2Freqs.get(id) - 1);
			if (site2Freqs.get(id) == 0)
				site2Freqs.remove(id);
			t.setSite(site);
			site.addTuple(t);
		}
		
	}
	
	public void randomNodeDistributionUsingBarabasiAlbertModel(int numOwners, int l, int numTries) {
		int tuplesIndex = 0;
		// make sure that each site has at least l tuples
		for (int i = 0; i < numOwners; i++)
		{
			Site s = SitesFactory.getSite(i);
			for (int j = 0; j < l; j++) 
				this.tuples.get(tuplesIndex++).setSite(s);
		}
		
		Map<Site, Integer> ownersFreq = Tools.getSitesFrequencies(this.tuples);
		for (int i = tuplesIndex; i < this.tuples.size(); i++){
			Site site = SitesFactory.getSite(randomUsingBAmodel(ownersFreq));
			this.tuples.get(i).setSite(site);
			ownersFreq.put(site, ownersFreq.get(site) + 1);
		}
		
		for (int aTry = 1; aTry < numTries; aTry++)
			for (int i = 0; i < this.tuples.size(); i++){
				Site prevSite = this.tuples.get(i).getSite();
				Site site = SitesFactory.getSite(randomUsingBAmodel(ownersFreq));
				this.tuples.get(i).setSite(site);
				ownersFreq.put(site, ownersFreq.get(site) + 1);
				ownersFreq.put(prevSite, ownersFreq.get(prevSite) - 1);
			}
	}

	private static int randomUsingBAmodel(Map<Site, Integer> ownersFreq) {
		double randomNumber = Math.random();
		double sum = 0.0;
		double sumPrev = 0.0;
		for (Integer i : ownersFreq.values())
			sum += i;
		
		double[] cumulativeDistribution = new double[ownersFreq.size()];
		for (int siteID = 0; siteID < ownersFreq.size(); siteID++)
		{
			int freq = ownersFreq.get(SitesFactory.getSite(siteID));
			sumPrev += freq / sum;
			cumulativeDistribution[siteID] = sumPrev;
		}

		int i = 0;
		while (i < cumulativeDistribution.length && cumulativeDistribution[i] < randomNumber)
			i++;
		
		return i;
	}
	
	/**
	 * Read tuples from the CSV file
	 * @param fileName name of the file
	 * @param withSiteInfo if true, then last column defines the date provider ID
	 * @return tuples read from the file
	 * @throws IOException
	 */
	public static RawTuples readTuples(String fileName, boolean withSiteInfo) throws IOException {
		List<Tuple> tuples = new ArrayList<Tuple>();
		BufferedReader reader = new BufferedReader(new FileReader(fileName));
		String line;
		while ((line = reader.readLine()) != null) {
			String[] attrs = line.replaceAll(" ", "").split(",");
			int numAttributes = (withSiteInfo ? attrs.length - 1 : attrs.length);
			int[] vals = new int[numAttributes];
			for (int i = 0; i < vals.length; i++) {
				vals[i] = Integer.parseInt(attrs[i]);
			}
			Site site = null;
			if (withSiteInfo) {
				site = SitesFactory.getSite(Integer.parseInt(attrs[attrs.length - 1]));
			}
			tuples.add(new Tuple(vals, site));
		}
		reader.close();
		
		RawTuples stats = new RawTuples(tuples);
		
		return stats;
	}

	public static void saveTuples(List<Tuple> tuples, String fileName, boolean withSiteInfo) throws IOException {
		BufferedWriter writer = new BufferedWriter(new FileWriter(fileName));
		
		for (Tuple t : tuples) {
			// save QIDs
			for (int i = 0; i < t.getData().length - 1; i++)
				writer.write(t.getData(i) + ",");
			
			// save sensitive data
			writer.write("" + t.getSensitiveData());
			
			if (withSiteInfo)
				writer.write("," + t.getSiteID());
			
			writer.newLine();
		}
		
		writer.close();
		
	}
	
	public int getNumQIDAttributes() {
		return this.getNumAllAttributes() - 1;
	}

	public void distributeTuples(TuplesGroup groups)
	{
		for (Tuple t : this.tuples)
		{
			TuplesGroup group = groups;
			while (group.hasSubGroups())
			{
				if (group.getLeftChild().canContainTuple(t))
					group = group.getLeftChild();
				else
					group = group.getRightChild();
			}
			group.addTuple(t);
		}
	}

	public void sampleTuples(int sampleSize, boolean shuffle)
	{
		if (sampleSize >= this.tuples.size())
			return;

		if (shuffle) {
			Collections.shuffle(this.tuples);
			Collections.shuffle(this.tuples);
		}

		List<Tuple> sampledTuples = new ArrayList<Tuple>(sampleSize);
		for (int i = 0; i < sampleSize; i++) {
			int randomIndex = (int)(Math.random() * this.tuples.size());
			Tuple t = this.tuples.remove(randomIndex);
			sampledTuples.add(t);
		}
		
		this.tuples.clear();
		this.tuples.addAll(sampledTuples);
		this.computeStatistics();
	}

}
