package edu.emory.mathcs;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import edu.emory.mathcs.privacy.L_diversity;
import edu.emory.mathcs.privacy.PrivacyChecker;

public class Site implements Comparable<Site> {

	private int siteID;
	private ArrayList<Tuple> tuples;
	
	public Site(int id)
	{
		this.siteID = id;
		this.tuples = new ArrayList<Tuple>();
	}
	
	public void addTuple(Tuple t)
	{
		this.tuples.add(t);
	}

	public void removeTuple(Tuple tuple)
	{
		this.tuples.remove(tuple);
	}
	
	// n > 0
	/**
	 * @return 1st: nth value, 2nd: # tuples with attribute 1st which should be in the set, 3rd: site ID
	 */
	public List<Integer> getNthSecureValueGT(double value, int n, final int attributeNum)
	{
		ArrayList<Tuple> tuples = Tools.getUncorruptedTuples(this.tuples);
		
		Collections.sort(tuples, new Comparator<Tuple>()
				{
					public int compare(Tuple o1, Tuple o2) {
						return o1.getData(attributeNum) - o2.getData(attributeNum);
				}
			}
		);

		//TODO change to binary search!
		int firstIndex = 0;
		while (firstIndex < tuples.size() 
				&& (tuples.get(firstIndex).isCorrupted() || tuples.get(firstIndex).getData(attributeNum) < value))
			firstIndex++;
		
		int lastIndex = firstIndex;
		int counter = 0;
		while(counter < n && lastIndex < tuples.size())
		{
			if (tuples.get(lastIndex).isCorrupted() == false)
				counter++;
			
			lastIndex++;
		}
		
		if (lastIndex == tuples.size())
			return null;

		// check if there are at least n extra tuples
		int veryLastIndex = lastIndex;
		counter = 0;
		while(counter < n && veryLastIndex < tuples.size())
		{
			if (tuples.get(veryLastIndex).isCorrupted() == false)
				counter++;
			
			veryLastIndex++;
		}
		
		if (veryLastIndex == tuples.size())
			return null;

		veryLastIndex = lastIndex;
		while (veryLastIndex >= firstIndex 
				&& (tuples.get(veryLastIndex).isCorrupted() || tuples.get(lastIndex).getData(attributeNum) == tuples.get(veryLastIndex).getData(attributeNum)))
			veryLastIndex--;
		
		return Arrays.asList(new Integer[] {tuples.get(lastIndex).getData(attributeNum), lastIndex - veryLastIndex, this.getID()});
	}
	
	@Override
	public int compareTo(Site o) {
		return this.siteID - o.siteID;
	}
	
	public int getNumTuples(){
		return this.tuples.size();
	}

//	public List<? extends Tuple> getTuples(final int attributeID, Integer minValue, Integer maxValue) {
//		
//		Collections.sort(this.tuples, new Comparator<Tuple>()
//				{
//					public int compare(Tuple o1, Tuple o2) {
//						return o1.getData(attributeID) - o2.getData(attributeID);
//				}
//			}
//		);
//
//		int firstIndex = 0;
//		while (firstIndex < this.tuples.size() && this.tuples.get(firstIndex).getData(attributeID) < minValue)
//			firstIndex++;
//
//		int lastIndex = firstIndex;
//		while (lastIndex < this.tuples.size() && this.tuples.get(lastIndex).getData(attributeID) <= maxValue)
//			lastIndex++;
//		
//		return this.tuples.subList(firstIndex, lastIndex);
//	}
	
	public int getID()
	{
		return this.siteID;
	}
	
	public String toString(){
		return this.getID() + "(" + this.getNumTuples() + ")";
	}
	
	public boolean isTriMdiverse(List<Tuple> tuples, final int m, final int attributeID)
	{
		if (tuples.size() < 2 * m)
			return false;
		
		Collections.sort(tuples, new Comparator<Tuple>(){
			public int compare(Tuple o1, Tuple o2)
			{
				return o1.getData(attributeID) - o1.getData(attributeID);
			}
		});
		
		int l1_numTuples = Site.getPrivNumTuples(tuples, new L_diversity(m));
		
		if (l1_numTuples < 0 || l1_numTuples > tuples.size())
			return false;
		
		List<Tuple> tuples_l1 = tuples.subList(l1_numTuples, tuples.size());
		return this.isBiMdiverse(tuples_l1, m, attributeID);
	}

	public static int getBiPrivNumTuples(final List<Tuple> tuples, final PrivacyChecker checker)
	{
		int l1_numTuples = Site.getPrivNumTuples(tuples, checker);
		Collections.reverse(tuples);

		int ll_numTuples = Site.getPrivNumTuples(tuples, checker);
		Collections.reverse(tuples);

		if (l1_numTuples + ll_numTuples <= tuples.size())
			return tuples.size();
		
		return l1_numTuples;
	}
	
	public static int getPrivNumTuples(final List<Tuple> tuples, final PrivacyChecker checker)
	{
		TuplesGroup tg = null;
		
		int begin = 0;
		int end = tuples.size();
		while (begin < end)
		{
			tg = new TuplesGroup(tuples.subList(0, (begin + end) / 2), null);
			if (checker.isAnonymized(tg) == false)
			{
				if (begin == end - 1)
				{
					begin++;
					tg = new TuplesGroup(tuples.subList(0, (begin + end) / 2), null);
				}
				else
					begin = (begin + end) / 2;
			}
			else
			{
				if (begin == end - 1)
					end--;
				else
					end = (begin + end) / 2;
			}

		}
		
		return (begin + end) / 2;
	}

	public boolean isBiMdiverse(List<Tuple> tuples, int minNumTuplesInGroup, int attributeID)
	{
		if (tuples.size() < 2 * minNumTuplesInGroup)
			return false;

		TuplesGroup tg = new TuplesGroup(null, null);
		tg.setTuplesGroups(new TuplesGroup(null, null), new TuplesGroup(null, null));
		
		return  (getMdivIntersectionBothExclusiveRange(tuples, minNumTuplesInGroup, attributeID, tg).isEmpty() == false);
	}

	public IntAttributeRange getMdivIntersectionBothExclusiveRange(List<Tuple> tuples, int m, final int attributeID, TuplesGroup splittedTuples)
	{
		Collections.sort(tuples, new Comparator<Tuple>(){
			@Override
			public int compare(Tuple arg0, Tuple arg1)
			{
				return arg0.getData(attributeID) - arg1.getData(attributeID);
			}
		});
		int l1_numTuples = Site.getPrivNumTuples(tuples, new L_diversity(m));
		Collections.reverse(tuples);

		int ll_numTuples = Site.getPrivNumTuples(tuples, new L_diversity(m));
		Collections.reverse(tuples);

		IntAttributeRange range = new IntAttributeRange(Integer.MAX_VALUE, Integer.MIN_VALUE);
		if (l1_numTuples + ll_numTuples <= tuples.size())
		{
			splittedTuples.getLeftChild().addTuples(tuples.subList(0, l1_numTuples));
			splittedTuples.addTuples(tuples.subList(l1_numTuples, tuples.size() - ll_numTuples));
			splittedTuples.getRightChild().addTuples(tuples.subList(tuples.size() - ll_numTuples, tuples.size()));
			
			range.setMinMax(tuples.get(l1_numTuples).getData(attributeID), tuples.get(tuples.size() - ll_numTuples - 1).getData(attributeID));
		}
		
		return range; 
	}

	public static Set<Tuple> getTuples(Set<Site> sites, final Map<Site, List<Tuple>> site2tuples) {
		Set<Tuple> splits = new HashSet<Tuple>();
		
		for (Site s : site2tuples.keySet()) {
			if (sites.contains(s)) {
				splits.addAll(site2tuples.get(s));
			}
		}
		
		return splits;
	}
	
	public static Set<Tuple> [] splitTuples(Set<Site> sites, final Map<Site, List<Tuple>> site2tuples) {
			Set<Tuple> [] splits = new HashSet[2];
			
			splits[0] = new HashSet<Tuple>();
			splits[1] = new HashSet<Tuple>();
			
			for (Site s : site2tuples.keySet()) {
				if (sites.contains(s)) {
					splits[0].addAll(site2tuples.get(s));
				}
				else {
					splits[1].addAll(site2tuples.get(s));
				}
			}
			
			return splits;
		}
		/*
		private Map<Set<Site>, Integer> getUnsureAnonymityGroups(final Map<Site, List<Tuple>> site2tuples, Map<Set<Site>, Integer> unsureAnonymityGroups, Set<Site> expandingSites)
		{
			// build groups
			if (unsureAnonymityGroups.isEmpty()) {
				Iterator<Site> siteIt = expandingSites.iterator();
				while(siteIt.hasNext()) {
					Set<Site> sites = new TreeSet<Site>();
					Site s = siteIt.next();
					sites.add(s);
					Set<Tuple> [] tss = DmChecker.splitSites(sites, site2tuples);
					TuplesGroup ts0 = new TuplesGroup(tss[0], null);
					TuplesGroup ts1 = new TuplesGroup(tss[1], null);
					
					// ts0 is not private - supersets may not be private as well
					// ts0 is unsure
					if (this.checker.isAnonymized(ts0) == false) 
						unsureAnonymityGroups.put(sites, s.getID());
					else
						siteIt.remove(); // ts0 is private - shouldn't be used to expand sets
					
					if (this.checker.isAnonymized(ts1) == false) {
						unsureAnonymityGroups.clear();
						return unsureAnonymityGroups;
					}
				}
			} else {
	
				Map<Set<Site>, Integer> newGroups = new HashMap<Set<Site>, Integer>();
				for (Set<Site> sites : unsureAnonymityGroups.keySet()) {
					for (Site s : expandingSites) {
						// To avoid necessary check for duplicated sites
						Integer maxSetSiteID = unsureAnonymityGroups.get(sites);
						if (maxSetSiteID != null && maxSetSiteID >= s.getID())
							continue;
						
						Set<Site> newSites = new TreeSet<Site>(sites);
						newSites.add(s);
						List<Tuple> ts = new ArrayList<Tuple>();
						for (Site newSite : newSites)
							ts.addAll(site2tuples.get(newSite));
						Set<Tuple> [] tss = DmChecker.splitSites(sites, site2tuples);
						TuplesGroup ts0 = new TuplesGroup(tss[0], null);
						TuplesGroup ts1 = new TuplesGroup(tss[1], null);
						// s has the greatest ID among all sites in this set
						if (this.checker.isAnonymized(ts0) == false)
							newGroups.put(newSites, s.getID());
	
						if (this.checker.isAnonymized(ts1) == false) {
							unsureAnonymityGroups.clear();
							return unsureAnonymityGroups;
						}
					}
					
				}
				 return newGroups;
			}
			
			return unsureAnonymityGroups;
		}
	*/
}
