package edu.emory.mathcs.anonymizers;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.TreeMap;

import edu.emory.mathcs.CombinationsGenerator;
import edu.emory.mathcs.Site;
import edu.emory.mathcs.Tools;
import edu.emory.mathcs.Tuple;
import edu.emory.mathcs.TuplesGroup;
import edu.emory.mathcs.privacy.AndCheckers;
import edu.emory.mathcs.privacy.K_anonymity;
import edu.emory.mathcs.privacy.PrivacyChecker;

public class AdaptiveMPrivacySplitter implements Anonymizer
{
	private static enum SPLIT_SITE {SORT_VALUES};
	private int minNumTuplesInGroup;
	private PrivacyChecker checker;
	private long lastPrivateSplitsNum;
	private long lastSplitsNum;
	
	public AdaptiveMPrivacySplitter(int minNumTuplesInGroup, PrivacyChecker checker)
	{
		this.minNumTuplesInGroup = minNumTuplesInGroup;
		this.checker = checker;
		this.lastPrivateSplitsNum = 0;
		this.lastSplitsNum = 0;
	}
	
	public void splitGroup(TuplesGroup group) {
		rawSplitGroup(group);
		group.reduceRedundantGroups();
	}
	
	private void rawSplitGroup(TuplesGroup group) {
		
		final int numAttributes = group.getNumAttributes();
		Stack<TuplesGroup> groupsToConsider = new Stack<TuplesGroup>();
		groupsToConsider.push(group);

		while (groupsToConsider.size() > 0)
		{
			group = groupsToConsider.pop();
			
			if (group.getNumTuples() < 2 * this.minNumTuplesInGroup)
				continue;

			// Medians and attribute ranges
			double[] median = new double[numAttributes];
			for (int i = 0; i < median.length; i++)
				median[i] = Tools.findMedian(i, group.getData());

			// Possible splits
			List<TuplesGroup[]> groups = new ArrayList<TuplesGroup[]>(numAttributes);
			List<Double> splittingMeasureValues = new ArrayList<Double>();
			TuplesGroup[] splitGroup1 = null;
			List<TuplesGroup[]> splitGroup2 = null;
			for (int i = 0; i < median.length; i++)
			{
				splitGroup1 = group.split(i, median[i], this.minNumTuplesInGroup, null);
				
				this.lastSplitsNum++;
				boolean isSplitGroup10Private = this.checker.isAnonymized(splitGroup1[0]);
				boolean isSplitGroup11Private = this.checker.isAnonymized(splitGroup1[1]);
				
				if (isSplitGroup10Private && isSplitGroup11Private)
				{
					this.lastPrivateSplitsNum++;
					groups.add(splitGroup1);
				}
			}

			splitGroup2 = this.splitSites(group, SPLIT_SITE.SORT_VALUES);
			if (splitGroup2 != null)
			{
				for (int i = 0; i < splitGroup2.size(); i++)
				{
					TuplesGroup[] sg = splitGroup2.get(i);
					if (sg == null)
						continue;
					
					boolean issg0Private = this.checker.isAnonymized(sg[0]);
					boolean issg1Private = this.checker.isAnonymized(sg[1]);
					if (issg0Private && issg1Private)
					{
						this.lastPrivateSplitsNum++;
						groups.add(sg);
					}
				}
			}

			for (int i = 0; i < groups.size(); i++)
			{
				TuplesGroup[] sg = groups.get(i);

				if (sg == null)
					continue;

				double leftMeasureValue = this.checker.getPrivacyFitness(sg[0]);
				double rightMeasureValue = this.checker.getPrivacyFitness(sg[1]);
				splittingMeasureValues.add(Math.min(leftMeasureValue, rightMeasureValue));

			}

			// Now I have rangesLen, maxRange, minNodes, min, max
			int best = -1;
			double bestScore = -1;
			for (int i = 0; i < groups.size(); i++)
			{
				if (groups.get(i) == null)
					continue;

				double score = splittingMeasureValues.get(i);
				
				if (score > bestScore)
				{
					bestScore = score;
					best = i;
				}
			}

			if (best != -1)
			{
				group.setTuplesGroups(groups.get(best)[0], groups.get(best)[1]);
				groupsToConsider.push(groups.get(best)[0]);
				groupsToConsider.push(groups.get(best)[1]);
			}
		}
		
		System.out.println("MNUM: " + this.lastSplitsNum + " splits, with " + this.lastPrivateSplitsNum + " private.");
	}

	public double getLastPrivateSplitEfficiency() {
		return (double) this.lastPrivateSplitsNum / (double) this.lastSplitsNum;
	}
	
	private boolean isSitePrivate(TuplesGroup tuples, PrivacyChecker siteChecker)
	{
		boolean isPrivate = true;
		Map<Site, List<Tuple>> site2tuples = tuples.getTuplesSplittedToSites();
		
		for (Site site : site2tuples.keySet())
		{
			isPrivate = isPrivate && siteChecker.isAnonymized(new TuplesGroup(site2tuples.get(site)));
			if (isPrivate == false)
				break;
		}
		
		return isPrivate;
	}

	private List<TuplesGroup[]> splitSites(TuplesGroup group, SPLIT_SITE sortValues)
	{
		TuplesGroup[] groups = new TuplesGroup[2];
		List<Tuple> list0 = new ArrayList<Tuple>();
		List<Tuple> list1 = new ArrayList<Tuple>();
		
		Map<Site, List<Tuple>> site2tuples = group.getTuplesSplittedToSites();

		if (sortValues == SPLIT_SITE.SORT_VALUES) {
			
			// get attribute values of first unnecessary tuple in the site 
			final Map<Site, Double> site2Values = new TreeMap<Site, Double>();
			
			if (site2Values.size() == 1)
				return null;
			
			int numAllTuples = 0;
			for (Site site : site2tuples.keySet()) {
				List<Tuple> ts = site2tuples.get(site);
//				site2Values.put(site, this.checker.getPrivacyFitness(new TuplesGroup(ts, null, null)));
//				site2Values.put(site, ts.size());
				numAllTuples += ts.size();
			}
			
			// sort sites by value
			List<Site> sites = new ArrayList<Site>(site2tuples.keySet());
// To test if sorting is needed.
//			Collections.sort(sites, new Comparator<Site>(){
//				public int compare(Site o1, Site o2)
//				{
//					return (int)Math.signum(site2Values.get(o2) - site2Values.get(o1));
//				}
//			});
//			
			// smart
			int i = 0;
			while (i < sites.size() && (list1.size() < numAllTuples / 2))
			{
				List<Tuple> ts = site2tuples.get(sites.get(i++));
				list1.addAll(ts);
			}
			
			for (; i < sites.size(); i++)
				list0.addAll(site2tuples.get(sites.get(i)));
		}
		
		this.lastSplitsNum++;
		// to speed up for k-anonymity case(s)
		if (list0.size() < this.minNumTuplesInGroup || list1.size() < this.minNumTuplesInGroup)
			return null;
		
		groups[0] = new TuplesGroup(list1);
		groups[1] = new TuplesGroup(list0);
		List<TuplesGroup[]> tg = new ArrayList<TuplesGroup[]>(1);
		tg.add(groups);
		
		return tg;
	}
	
	private List<TuplesGroup[]> splitSitesBruteForce(TuplesGroup group)
	{
		List<TuplesGroup[]> groups = new ArrayList<TuplesGroup[]>();
		Map<Site, List<Tuple>> site2tuples = group.getTuplesSplittedToSites();
		final int totalNumSites = site2tuples.size();
		final int totalNumTuples = group.getNumTuples();
		CombinationsGenerator<Site> cg = new CombinationsGenerator<Site>(site2tuples.keySet());
		
		for (int numSites = 1; numSites <= totalNumSites /2; numSites++)
		{
		
			cg.setCardinality(numSites);
			
			while (cg.hasNext()) {
				Set<Site> sites = cg.next();
				
				int numTuples = 0;
				for (Site s : sites)
					numTuples += site2tuples.get(s).size();
				
				if (numTuples < this.minNumTuplesInGroup || totalNumTuples - numTuples < this.minNumTuplesInGroup)
					continue;
				
				TuplesGroup[] ts = new TuplesGroup[2];
				Set<Tuple>[] tss = Site.splitTuples(sites, site2tuples);

				ts[0] = new TuplesGroup(tss[0]/*, group.getStats()*/);
				ts[1] = new TuplesGroup(tss[1]/*, group.getStats()*/);
				
				if (this.checker.isAnonymized(ts[1]) == false) {
					cg.forbidCurrentAndAllSuper();
					continue;
				}
				
				if (this.checker.isAnonymized(ts[0]) == false)
					continue;
				
				groups.add(ts);
			}
		}
		return groups;
	}

	public String getOutFileName()
	{
		return "KAttributeMondrian";
	}

	@Override
	public PrivacyChecker getPrivacyChecker()
	{
		return new AndCheckers.AndCheckersBuilder()
				.addChecker(this.checker, 1.0)
				.addChecker(new K_anonymity(this.minNumTuplesInGroup), 1.0)
				.build("K_mondrian_splitter_checker");
	}
}
