package edu.emory.mathcs.anonymizers;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import edu.emory.mathcs.Site;
import edu.emory.mathcs.Tools;
import edu.emory.mathcs.Tuple;
import edu.emory.mathcs.TuplesGroup;
import edu.emory.mathcs.privacy.AndCheckers;
import edu.emory.mathcs.privacy.K_anonymity;
import edu.emory.mathcs.privacy.PrivacyChecker;

public class KAttrMondrianSplitterMondrian implements Anonymizer
{
	private static enum SPLIT_SITE {SORT_VALUES};
	private int minNumTuplesInGroup;
	private PrivacyChecker checker;
	private double lastPrivateSplitEfficiency;
	
	public KAttrMondrianSplitterMondrian(int minNumTuplesInGroup, PrivacyChecker checker)
	{
		this.minNumTuplesInGroup = minNumTuplesInGroup;
		this.checker = checker;
		this.lastPrivateSplitEfficiency = 0;
	}
	
	public void splitGroup(TuplesGroup group) {
		rawSplitGroup(group);
		group.reduceRedundantGroups();
	}
	
	private void rawSplitGroup(TuplesGroup group) {
		
		long numSplitGroups = 0;
		long numPrivateSplitGroups = 0;
		final int numAttributes = group.getNumAttributes();
		Stack<TuplesGroup> groupsToConsider = new Stack<TuplesGroup>();
		groupsToConsider.push(group);

		while (groupsToConsider.size() > 0)
		{
			group = groupsToConsider.pop();
			
			if (group.getNumTuples() < 2 * this.minNumTuplesInGroup)
				continue;

			// Medians and attribute ranges
			double[] median = new double[numAttributes];
			for (int i = 0; i < median.length; i++)
				median[i] = Tools.findMedian(i, group.getData());

			// Possible splits
			List<TuplesGroup[]> groups = new ArrayList<TuplesGroup[]>(numAttributes);
			List<Double> splittingMeasureValues = new ArrayList<Double>();
			TuplesGroup[] splitGroup1 = null;
			for (int i = 0; i < median.length; i++)
			{
				splitGroup1 = group.split(i, median[i], this.minNumTuplesInGroup, null);
				
				numSplitGroups++;
				if (this.checker.isAnonymized(splitGroup1[0]) 
					&& this.checker.isAnonymized(splitGroup1[1]))
				{
					numPrivateSplitGroups++;
					groups.add(splitGroup1);
				}
			}

			for (int i = 0; i < groups.size(); i++)
			{
				TuplesGroup[] sg = groups.get(i);

				if (sg == null)
					continue;

				double leftMeasureValue = this.checker.getPrivacyFitness(sg[0]);
				double rightMeasureValue = this.checker.getPrivacyFitness(sg[1]);
				splittingMeasureValues.add(Math.min(leftMeasureValue, rightMeasureValue));

			}

			// Now I have rangesLen, maxRange, minNodes, min, max
			int best = -1;
			double bestScore = -1;
			for (int i = 0; i < groups.size(); i++)
			{
				if (groups.get(i) == null)
					continue;

				double score = splittingMeasureValues.get(i);
				
				if (score > bestScore)
				{
					bestScore = score;
					best = i;
				}
			}

			if (best != -1)
			{
				group.setTuplesGroups(groups.get(best)[0], groups.get(best)[1]);
				groupsToConsider.push(groups.get(best)[0]);
				groupsToConsider.push(groups.get(best)[1]);
			}
		}
		
		lastPrivateSplitEfficiency = (double) numPrivateSplitGroups / (double) numSplitGroups;
		System.out.println("MNUM-Mondrian: " + numSplitGroups + " splits, with " + numPrivateSplitGroups + " private.");
	}
	
	public double getLastPrivateSplitEfficiency() {
		return lastPrivateSplitEfficiency;
	}
	
	private boolean isSitePrivate(TuplesGroup tuples, PrivacyChecker siteChecker)
	{
		boolean isPrivate = true;
		Map<Site, List<Tuple>> site2tuples = tuples.getTuplesSplittedToSites();
		
		for (Site site : site2tuples.keySet())
		{
			isPrivate = isPrivate && siteChecker.isAnonymized(new TuplesGroup(site2tuples.get(site)));
			if (isPrivate == false)
				break;
		}
		
		return isPrivate;
	}

	public String getOutFileName()
	{
		return "KAttributeMondrian";
	}

	@Override
	public PrivacyChecker getPrivacyChecker()
	{
		return new AndCheckers.AndCheckersBuilder()
				.addChecker(this.checker, 1.0)
				.addChecker(new K_anonymity(this.minNumTuplesInGroup), 1.0)
				.build("K_mondrian_splitter_checker");
	}
}
