package edu.emory.mathcs.anonymizers;

import java.util.Stack;

import edu.emory.mathcs.Tools;
import edu.emory.mathcs.TuplesGroup;
import edu.emory.mathcs.privacy.PrivacyChecker;

public class KHeuristicMondrianSplitter implements Anonymizer {

	private int k;
	private PrivacyChecker checker;
	private double lastPrivateSplitEfficiency;
	
	public KHeuristicMondrianSplitter(int k, PrivacyChecker checker) {
		this.k = k;
		this.checker = checker;
		this.lastPrivateSplitEfficiency = 0;
	}
	
	public void splitGroup(TuplesGroup group) {
		rawSplitGroup(group);
//		group.reduceRedundantGroups();
	}
	
	public void rawSplitGroup(TuplesGroup groupToSplit) {
		
		long numSplitGroups = 0;
		long numPrivateSplitGroups = 0;
		final int numAttributes = groupToSplit.getNumAttributes();
		Stack<TuplesGroup> groupsToConsider = new Stack<TuplesGroup>();
		groupsToConsider.push(groupToSplit);

		while (groupsToConsider.size() > 0) {
			
			TuplesGroup group = groupsToConsider.pop();
			
			if (group.getNumTuples() < 2 * this.k) {
				continue;
			}
			
			//Medians and attribute ranges
			double[] median = new double[numAttributes];
			for (int i = 0; i < median.length; i++) {
				median[i] = Tools.findMedian(i, group.getData());
			}
			
			//Possible splits
			TuplesGroup[][] groups = new TuplesGroup[numAttributes][];
			double[] splittingMeasureValues = new double[groups.length];
			for (int i = 0; i < median.length; i++) 
			{
				groups[i] = group.split(i, median[i], this.k, null);
	
				numSplitGroups++;
				if (this.checker.isAnonymized(groups[i][0]) && this.checker.isAnonymized(groups[i][1]))
				{
					numPrivateSplitGroups++;
					double leftMeasureValue = this.checker.getPrivacyFitness(groups[i][0]);
					double rightMeasureValue = this.checker.getPrivacyFitness(groups[i][1]);
					splittingMeasureValues[i] = Math.min(leftMeasureValue, rightMeasureValue);
				}
				else
				{
					groups[i] = null;
					splittingMeasureValues[i] = -1;
				}
			}
			
			//Now I have everything
			int best = -1;
			double bestScore = -1;
			double score = -1;
			for (int i = 0; i < splittingMeasureValues.length; i++) {
				if (groups[i] == null)
					continue;
	
				score = splittingMeasureValues[i];
				
				if (score > bestScore) {
					bestScore = score;
					best = i;
				}
			}
			
			if (best != -1) {
				group.setTuplesGroups(groups[best][0], groups[best][1]);
				groupsToConsider.push(groups[best][0]);
				groupsToConsider.push(groups[best][1]);
			}
		}
		
		lastPrivateSplitEfficiency = (double) numPrivateSplitGroups / (double) numSplitGroups;
		System.out.println("Mondrian: " + numSplitGroups + " splits, with " + numPrivateSplitGroups + " private.");
	}

	public double getLastPrivateSplitEfficiency() {
		return lastPrivateSplitEfficiency;
	}
	
	public String getOutFileName() {
		return "KHeuristicMondrian";
	}
	
	@Override
	public PrivacyChecker getPrivacyChecker()
	{
		return this.checker;
	}
}
