package edu.emory.mathcs.anonymizers;

import edu.emory.mathcs.Tools;
import edu.emory.mathcs.TuplesGroup;
import edu.emory.mathcs.measures.Measure;
import edu.emory.mathcs.privacy.AndCheckers;
import edu.emory.mathcs.privacy.K_anonymity;
import edu.emory.mathcs.privacy.PrivacyChecker;

public class KLSiteMondrianSplitter implements Anonymizer {

	private int k;
	private int l;
	private double alpha = 0.5;
	private PrivacyChecker checker;
	private Measure splittingMeasure;
	
	public KLSiteMondrianSplitter(int k, int l, double alpha, Measure splittingMeasure, PrivacyChecker checker) {
		this.k = k;
		this.l = l;
		this.alpha = alpha;
		this.checker = checker;
		this.splittingMeasure = splittingMeasure;
	}
	
	public void splitGroup(TuplesGroup group) {
		
		if (group.getNumTuples() < 2 * k) {
			return;
		}
		
		//Medians and attribute ranges
		double[] median = new double[group.getNumAttributes()];
		for (int i = 0; i < median.length; i++) {
			median[i] = Tools.findMedian(i, group.getData());
		}
		int[] rangeLen = new int[median.length];
		rangeLen[0] = group.getAttributeRange(0).getMax() - group.getAttributeRange(0).getMin();
		int maxRange = rangeLen[0];
		int minRange = maxRange;
		for (int i = 1; i < rangeLen.length; i++) {
			rangeLen[i] = group.getAttributeRange(i).getMax() - group.getAttributeRange(i).getMin();
			if (rangeLen[i] > maxRange) {
				maxRange = rangeLen[i];
			}
			if (rangeLen[i] < minRange) {
				minRange = rangeLen[i];
			}
		}
		
		//Possible splits
		TuplesGroup[][] groups = new TuplesGroup[median.length][];
		double[] splittingMeasureValues = new double[median.length];
		double minMeasureValue = Double.MAX_VALUE;
		double maxMeasureValue = 0;
		for (int i = 0; i < groups.length; i++) {
			groups[i] = group.split(i, median[i], this.k, null);
			
			double leftMeasureValue = this.splittingMeasure.getValue(groups[i][0]);
			double rightMeasureValue = this.splittingMeasure.getValue(groups[i][1]);
			splittingMeasureValues[i] = this.splittingMeasure.getCombinedMeasureValue(leftMeasureValue, rightMeasureValue);

			maxMeasureValue = Math.max(maxMeasureValue, splittingMeasureValues[i]);
			minMeasureValue = Math.min(minMeasureValue, splittingMeasureValues[i]);
		}
		
		
		//Now I have rangesLen, maxRange, minNodes, min, max
		int best = -1;
		double bestScore = -1;
		for (int i = 0; i < splittingMeasureValues.length; i++) {
			if ((groups[i][0].getCountOfDistinctNodes(false) < l)
					|| (groups[i][1].getCountOfDistinctNodes(false) < l)
					|| (this.checker.isAnonymized(groups[i][0]) == false) 
					|| (this.checker.isAnonymized(groups[i][1]) == false)
				)
			{ 
				continue;
			}
			double factorRange = getProjectedTo01(minRange, maxRange, rangeLen[i]);
			double factorDistribution = getProjectedTo01(minMeasureValue, maxMeasureValue, splittingMeasureValues[i]);
			double score = alpha * factorRange + (1 - alpha) * factorDistribution;
			if (this.splittingMeasure.getBetterScore(score, bestScore) != bestScore) {
				bestScore = score;
				best = i;
			}
		}
		
		if (best != -1) {
			group.setTuplesGroups(groups[best][0], groups[best][1]);
			splitGroup(groups[best][0]);
			splitGroup(groups[best][1]);
		}
		
	}
	
//	private double boundry(int l, int n, int s) {
//		return factor(n/2-l+1, n) - (l - 1) * factor(1, n) - s * factor(((n) / 2 - l - 1) / s * 2,  n) / 2;
//	}
	
//	private double factor(int cc, int n) {
//		return ((double)cc)/n * Math.log(((double)cc)/n);
//	}

//	public double entropy(Map<String, Map<Integer, Integer> > distribution) {
//		double entropy = 0;
//		double total;
//		
//		// for each group
//		for (Map<Integer, Integer> groupStats : distribution.values())
//		{
//			total = 0.0;
//			
//			for (Integer freq : groupStats.values())
//				total += freq;
//			
//			for (Integer freq : groupStats.values())
//			{
//				double pi = freq / (double) total;
//				if (freq == total)
//					pi = 1.0 / (double) total;
//								
//				if (pi != 0) {
//					entropy -= /*pi * */ Math.log(pi);
//				}
//			}
//		}
//		
//		return entropy;
//	}

	private double getProjectedTo01(double min, double max, double val) {
		return (val - min)/(max - min);
	}

	public String getOutFileName() {
		return "KLSiteMondrianSplitter";
	}
	
	@Override
	public PrivacyChecker getPrivacyChecker()
	{
		return new AndCheckers.AndCheckersBuilder()
			.addChecker(this.checker, 1.0)
			.addChecker(new K_anonymity(this.k), 1.0)
			.build("KL_site_mondrian_splitter_checker");
	}
}
