package edu.emory.mathcs.anonymizers;

import edu.emory.mathcs.Tools;
import edu.emory.mathcs.TuplesGroup;
import edu.emory.mathcs.privacy.AndCheckers;
import edu.emory.mathcs.privacy.K_anonymity;
import edu.emory.mathcs.privacy.L_diversity;
import edu.emory.mathcs.privacy.PrivacyChecker;

public class KLStdDevSplitter implements Anonymizer {
	
	private int k;
	private int l;
	private double alpha = 0.5;
	
	public KLStdDevSplitter(int k, int l, double alpha) {
		this.k = k;
		this.l = l;
		this.alpha = alpha;
	}
	
	public void splitGroup(TuplesGroup group) {
		//End condition
		if (group.getNumTuples() < 2 * k) {
			return;
		}
		
		//Medians and attribute ranges
		double[] median = new double[group.getNumAttributes()];
		for (int i = 0; i < median.length; i++) {
			median[i] = Tools.findMedian(i, group.getData());
		}
		int[] rangeLen = new int[median.length];
		int maxRange = 0;
		int minRange = Integer.MAX_VALUE;
		for (int i = 0; i < rangeLen.length; i++) {
			rangeLen[i] = group.getAttributeRange(i).getMax() - group.getAttributeRange(i).getMin();
			if (rangeLen[i] > maxRange) {
				maxRange = rangeLen[i];
			}
			if (rangeLen[i] < minRange) {
				minRange = rangeLen[i];
			}
		}
		
		//Possible splits
		TuplesGroup[][] groups = new TuplesGroup[median.length][];
		int[] minNodes = new int[median.length];
		int min = Integer.MAX_VALUE;
		int max = 0;
		double devMin = Integer.MAX_VALUE;
		double devMax = 0;
		double[] deviations = new double[median.length];
		for (int i = 0; i < groups.length; i++) {
			groups[i] = group.split(i, median[i], this.k, null);
			minNodes[i] = Math.min(groups[i][0].getCountOfDistinctNodes(false), groups[i][1].getCountOfDistinctNodes(false));
			if (min > minNodes[i]) {
				min = minNodes[i];
			}
			int lMax = Math.max(groups[i][0].getCountOfDistinctNodes(false), groups[i][1].getCountOfDistinctNodes(false));
			if (max < lMax) {
				max = lMax;
			}
			double localDevMin = Math.min(groups[i][0].getStdDeviationForNodesDistribution(), 
					groups[i][1].getStdDeviationForNodesDistribution());
			double localDevMax = Math.max(groups[i][0].getStdDeviationForNodesDistribution(), 
					groups[i][1].getStdDeviationForNodesDistribution());
			deviations[i] = localDevMin;
			if (localDevMin < devMin) {
				devMin = localDevMin;
			}
			if (localDevMax > devMax) {
				devMax = localDevMax;
			}
		}
		
		//Now I have rangesLen, maxRange, minNodes, min, max
		int best = -1;
		double bestScore = -1;
		for (int i = 0; i < minNodes.length; i++) {
			if (groups[i][0].getCountOfDistinctNodes(false) < l || groups[i][1].getCountOfDistinctNodes(false) < l) {
				continue;
			}
			double factorRange = getProjectedTo01(minRange, maxRange, rangeLen[i]);
			double factorDistribution = getProjectedTo01(devMin, devMax, deviations[i]);
			double score = alpha * factorRange + (1 - alpha) * (1-factorDistribution);
			if (score > bestScore) {
				bestScore = score;
				best = i;
			}
		}
		
		if (best != -1) {
			group.setTuplesGroups(groups[best][0], groups[best][1]);
			splitGroup(groups[best][0]);
			splitGroup(groups[best][1]);
		}
		
	}

	private double getProjectedTo01(double min, double max, double val) {
		return (val - min)/(max - min);
	}

	public String getOutFileName() {
		return "StdDev";
	}

	@Override
	public PrivacyChecker getPrivacyChecker()
	{
		return new AndCheckers.AndCheckersBuilder()
			.addChecker(new L_diversity(this.l), 1.0)
			.addChecker(new K_anonymity(this.k), 1.0)
			.build("KL_std_dev_splitter_checker");
	}
}
