package edu.emory.mathcs.anonymizers;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.TreeMap;
import java.util.TreeSet;

import edu.emory.mathcs.IntAttributeRange;
import edu.emory.mathcs.Site;
import edu.emory.mathcs.Tools;
import edu.emory.mathcs.Tuple;
import edu.emory.mathcs.TuplesGroup;
import edu.emory.mathcs.measures.Measure;
import edu.emory.mathcs.privacy.AndCheckers;
import edu.emory.mathcs.privacy.K_anonymity;
import edu.emory.mathcs.privacy.L_site_diversity;
import edu.emory.mathcs.privacy.PrivacyChecker;

public class KMMondrianSplitter implements Anonymizer
{
	private int k;
	private int m;
	private double alpha = 0.5;
	private PrivacyChecker checker;
	private Measure splittingMeasure;
	
	public KMMondrianSplitter(int k, int m, double alpha, Measure splittingMeasure, PrivacyChecker checker)
	{
		this.k = k;
		this.m = m;
	}
	
	@Override
	public String getOutFileName()
	{
		return "KMMondrianSplitter";
	}

	@Override
	public void splitGroup(TuplesGroup tuplesToSplit)
	{
		Stack<TuplesGroup> groupsToConsider = new Stack<TuplesGroup>();
		groupsToConsider.push(tuplesToSplit);

		TuplesGroup group;
		while (groupsToConsider.size() > 0)
		{
			group = groupsToConsider.pop();

			if (group.getNumTuples() < 2 * this.k)
			{
				return;
			}

			int[] rangeLen = new int[group.getNumAttributes()];
			rangeLen[0] = group.getAttributeRange(0).getMax() - group.getAttributeRange(0).getMin();
			int maxRange = rangeLen[0];
			int minRange = maxRange;
			for (int i = 1; i < rangeLen.length; i++)
			{
				rangeLen[i] = group.getAttributeRange(i).getMax() - group.getAttributeRange(i).getMin();
				if (rangeLen[i] > maxRange)
				{
					maxRange = rangeLen[i];
				}
				if (rangeLen[i] < minRange)
				{
					minRange = rangeLen[i];
				}
			}

			Map<Site, List<Tuple>> site2tuples = group.getTuplesSplittedToSites();
			double[] splittingMeasureValues = new double[group.getNumAttributes()];
			double minMeasureValue = Double.MAX_VALUE;
			double maxMeasureValue = 0;

			// chose the best attribute to split
			for (int attributeID = 0; attributeID < group.getNumAttributes(); attributeID++)
			{

				TuplesGroup splittedGroup = this.attributeSplitGroup(group, site2tuples, attributeID);

				if (splittedGroup == null || this.checker.isAnonymized(splittedGroup))
					continue;

				splittingMeasureValues[attributeID] = this.splittingMeasure.getValue(splittedGroup);
				maxMeasureValue = Math.max(maxMeasureValue, splittingMeasureValues[attributeID]);
				minMeasureValue = Math.min(minMeasureValue, splittingMeasureValues[attributeID]);
			}

			// Now I have everything to find the attribute for split
			int best = -1;
			double bestScore = -1;
			for (int i = 0; i < splittingMeasureValues.length; i++)
			{

				double factorRange = Tools.getProjectedTo01(minRange, maxRange, rangeLen[i]);
				double factorDistribution = Tools.getProjectedTo01(minMeasureValue, maxMeasureValue, splittingMeasureValues[i]);
				double score = this.alpha * factorRange + (1 - this.alpha) * factorDistribution;

				if (this.splittingMeasure.getBetterScore(score, bestScore) != bestScore)
				{
					bestScore = score;
					best = i;
				}
			}

			if (best != -1)
			{
				TuplesGroup splittedGroup = this.attributeSplitGroup(group, site2tuples, best);
				groupsToConsider.push(splittedGroup.getRightChild().getLeftChild());
				groupsToConsider.push(splittedGroup.getRightChild().getRightChild());
			}
		}

	}

	protected TuplesGroup attributeSplitGroup(TuplesGroup group, Map<Site, List<Tuple>> site2tuples, int attributeID)
	{
		// divide sites into 2 groups - each with different approach in splitting
		Set<Site> notSplittableSites = new TreeSet<Site>();
		int notSplittableSites_numTuples = 0;
		Set<Site> splittableSites = new TreeSet<Site>();
		int splittableSites_numTuples = 0;
		
		for (Site site : site2tuples.keySet())
		{
			List<Tuple> tuples = site2tuples.get(site);
			if (tuples == null)
				continue;

			if (tuples.size() < this.m)
				return null;

			if (site.isBiMdiverse(tuples, this.m, attributeID) == false)
			{
				notSplittableSites.add(site);
				notSplittableSites_numTuples += tuples.size();
				continue;
			}

			splittableSites.add(site);
			splittableSites_numTuples += tuples.size();
		}

		// checking if splitting it possible!
		if (splittableSites_numTuples < 2 * this.k)
			return null;

		// add tuples from not triMdiverse tuples
		Iterator<Site> sitesIterator = splittableSites.iterator();
		while (sitesIterator.hasNext() && notSplittableSites_numTuples < this.k && notSplittableSites_numTuples > 0)
		{
			Site site = sitesIterator.next();
			int numTuples = site2tuples.get(site).size();

			if (site.isTriMdiverse(site2tuples.get(site), m, attributeID) == false)
			{
				notSplittableSites.add(site);
				notSplittableSites_numTuples += numTuples;
				
				splittableSites_numTuples -= numTuples;
			}
		}
		splittableSites.removeAll(notSplittableSites);

		// TODO no 'tetris' for now
		if (splittableSites_numTuples < 2 * this.k)
			return null;

		List<Tuple> constLeftTuples = new ArrayList<Tuple>(this.k);
		List<Tuple> rightTuples = new ArrayList<Tuple>(this.k);
		for (Site site : site2tuples.keySet())
		{
			if (splittableSites.contains(site))
				rightTuples.addAll(site2tuples.get(site));
			else
				constLeftTuples.addAll(site2tuples.get(site));
		}

		
		IntAttributeRange common = new IntAttributeRange(Integer.MIN_VALUE, Integer.MAX_VALUE);
		Map<Site, IntAttributeRange> site2MdivIntersectionRange = new TreeMap<Site, IntAttributeRange>();
		TuplesGroup lowestValueTuples = new TuplesGroup(new ArrayList<Tuple>(), null);
		TuplesGroup higherValueTuples = new TuplesGroup(new ArrayList<Tuple>(), null);
		TuplesGroup medianTuples = new TuplesGroup(null, null);
		medianTuples.setTuplesGroups(lowestValueTuples, higherValueTuples);
		// get intersection of modified ranges 
		for (Site site : site2tuples.keySet())
		{
			IntAttributeRange range = site.getMdivIntersectionBothExclusiveRange(site2tuples.get(site), m, attributeID, medianTuples);
			site2MdivIntersectionRange.put(site, range);
			common = common.intersect(range);
		}
		
		if (common.isEmpty())
			common.reverseEnds();
		
		List<Tuple> commonTuples = new ArrayList<Tuple>();
		commonTuples.addAll(this.getTuplesWithinBothInclusiveRange(Arrays.asList(medianTuples.getData()), common, attributeID));

		//Medians and attribute ranges
		double median = Tools.findMedian(attributeID, commonTuples);

		// split remining tuples based on median
		for (Tuple t : medianTuples.getData())
		{
			if (t.getData(attributeID) < median)
				medianTuples.getLeftChild().addTuple(t);
			else if (t.getData(attributeID) > median)
				medianTuples.getRightChild().addTuple(t);
			else if (medianTuples.getLeftChild().getNumTuples() > medianTuples.getRightChild().getNumTuples())
				medianTuples.getRightChild().addTuple(t);
			else
				medianTuples.getLeftChild().addTuple(t);
		}
		
		medianTuples.clearTuples();
		
		return medianTuples;
	}

	private List<Tuple> getTuplesWithinBothInclusiveRange(List<Tuple> tuples, IntAttributeRange range, int attributeID)
	{
		List<Tuple> withinRange = new ArrayList<Tuple>();
		
		for (Tuple t : tuples)
		{
			if (range.isWithin(t.getData(attributeID)))
				withinRange.add(t);
		}
		
		return withinRange;
	}

	@Override
	public PrivacyChecker getPrivacyChecker()
	{
		return new AndCheckers.AndCheckersBuilder()
			.addChecker(this.checker, 1.0)
			.addChecker(new L_site_diversity(this.m), 1.0)
			.addChecker(new K_anonymity(this.k), 1.0)
			.build("KM_mondrian_splitter_checker");
	}
}
