package edu.emory.mathcs.anonymizers;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;
import java.util.TreeMap;

import edu.emory.mathcs.Site;
import edu.emory.mathcs.Tools;
import edu.emory.mathcs.Tuple;
import edu.emory.mathcs.TuplesGroup;
import edu.emory.mathcs.privacy.M_privacy;
import edu.emory.mathcs.privacy.PrivacyChecker;

public class KRealAttrMondrianSplitter implements Anonymizer
{
	private static enum SPLIT_SITE
	{
		SORT_VALUES
	};

	private int minNumTuplesInGroup;
	private M_privacy checker;
	private double lastPrivateSplitEfficiency;

	public KRealAttrMondrianSplitter(int minNumTuplesInGroup, M_privacy checker)
	{
		this.minNumTuplesInGroup = minNumTuplesInGroup;
		this.checker = checker;
		this.lastPrivateSplitEfficiency = 0.0;
	}

	public void splitGroup(TuplesGroup group)
	{
		long numSplitGroups = 0;
		long numPrivateSplitGroups = 0;
		final int numAttributes = group.getNumAttributes();
		Stack<TuplesGroup> groupsToConsider = new Stack<TuplesGroup>();
		groupsToConsider.push(group);
		
		while (groupsToConsider.size() > 0)
		{
			group = groupsToConsider.pop();
			
			if (group.getNumTuples() < 2 * this.minNumTuplesInGroup)
				continue;
	
			// Medians and attribute ranges
			double[] median = new double[numAttributes];
			for (int i = 0; i < median.length; i++)
			{
				median[i] = Tools.findMedian(i, group.getData());
			}
	
			// Possible splits
			TuplesGroup[][] groups = new TuplesGroup[median.length * 2][];
			double[] splittingMeasureValues = new double[groups.length];
			for (int i = 0; i < median.length; i++)
			{
				groups[2 * i] = group.split(i, median[i], this.minNumTuplesInGroup, null);
				numSplitGroups++;
	
				if (this.checker.isAnonymized(groups[2 * i][0]) == false 
						|| this.checker.isAnonymized(groups[2 * i][1]) == false)
					groups[2 * i] = null;
				else
					numPrivateSplitGroups++;
	
				Map<TuplesGroup, Boolean> group2certanity = this.splitSites(group, i, SPLIT_SITE.SORT_VALUES);
				numSplitGroups++;
				if (group2certanity != null && group2certanity.size() == 2)
				{
					groups[2 * i + 1] = new TuplesGroup[2];
					int j = 0;
					for (TuplesGroup tg : group2certanity.keySet())
					{
						groups[2 * i + 1][j++] = tg;
						if (group2certanity.get(tg) == false)
							group2certanity.put(tg, this.checker.isAnonymized(groups[2 * i + 1][0]));
					}
	
					if (group2certanity.get(groups[2 * i + 1][0]) == false
							|| group2certanity.get(groups[2 * i + 1][1]) == false)
						groups[2 * i + 1] = null;
					else
						numPrivateSplitGroups++;
				}
			}
			for (int i = 0; i < groups.length; i++)
			{
				if (groups[i] == null)
					continue;
	
				double leftMeasureValue = this.checker.getPrivacyFitness(groups[i][0]);
				double rightMeasureValue = this.checker.getPrivacyFitness(groups[i][1]);
				splittingMeasureValues[i] = Math.min(leftMeasureValue, rightMeasureValue);
			}
	
			// Now I have rangesLen, maxRange, minNodes, min, max
			int best = -1;
			double bestScore = -1;
			for (int i = 0; i < groups.length; i++)
			{
				if (groups[i] == null)
					continue;
				if (this.checker.isAnonymized(groups[i][0]) == false
						|| this.checker.isAnonymized(groups[i][1]) == false)
					continue;
	
				double score = splittingMeasureValues[i];
	
				if (score > bestScore)
				{
					bestScore = score;
					best = i;
				}
			}
	
			if (best != -1)
			{
				group.setTuplesGroups(groups[best][0], groups[best][1]);
				groupsToConsider.push(groups[best][0]);
				groupsToConsider.push(groups[best][1]);
			}
		}

		lastPrivateSplitEfficiency = (double) numPrivateSplitGroups / (double) numSplitGroups;
		System.out.println("MMIN1: " + numSplitGroups + " splits, with " + numPrivateSplitGroups + " private.");
	}

	private Map<TuplesGroup, Boolean> splitSites(TuplesGroup group, final int attributeID, SPLIT_SITE sortValues)
	{
		Map<TuplesGroup, Boolean> groups = new HashMap<TuplesGroup, Boolean>();
		List<Tuple> list0 = new ArrayList<Tuple>();
		boolean list0_private = false;
		List<Tuple> list1 = new ArrayList<Tuple>();
		boolean list1_private = false;

		Map<Site, List<Tuple>> site2tuples = group.getTuplesSplittedToSites();

		if (site2tuples.size() == 1)
			return null;

		// sort tuples for each site
		for (List<Tuple> tuples : site2tuples.values())
		{
			Collections.sort(tuples, new Comparator<Tuple>()
				{
					public int compare(Tuple arg0, Tuple arg1)
					{
						return arg0.getData(attributeID) - arg1.getData(attributeID);
					}
				});
		}

		if (sortValues == SPLIT_SITE.SORT_VALUES)
		{

			// get attribute values of first unnecessary tuple in the site
			final Map<Site, Integer> site2Values = new TreeMap<Site, Integer>();
			for (Site site : site2tuples.keySet())
			{
				List<Tuple> ts = site2tuples.get(site);
				int tupleNum = Site.getPrivNumTuples(ts, checker
						.getPrivacyChecker());

				int minPrivateSplitPoint = Integer.MAX_VALUE;
				if (tupleNum < ts.size())
					minPrivateSplitPoint = ts.get(tupleNum)
							.getData(attributeID);

				site2Values.put(site, minPrivateSplitPoint);
			}

			// sort sites by value
			List<Site> sites = new ArrayList<Site>(site2tuples.keySet());
			Collections.sort(sites, new Comparator<Site>()
				{
					public int compare(Site o1, Site o2)
					{
						return site2Values.get(o2) - site2Values.get(o1);
					}
				});

			// Median
			for (int i = 0; i < sites.size() / 2; i++)
				list1.addAll(site2tuples.get(sites.get(i)));
			for (int i = sites.size() / 2; i < sites.size(); i++)
				list0.addAll(site2tuples.get(sites.get(i)));

			Site medianSite = sites.get(sites.size() / 2 - 1);
			if (site2Values.get(medianSite) < Integer.MAX_VALUE)
				list1_private = true;
			Site lastSite = sites.get(sites.size() - 1);
			if (site2Values.get(lastSite) < Integer.MAX_VALUE)
				list0_private = true;

		}

		if (list0.size() < this.minNumTuplesInGroup || list1.size() < this.minNumTuplesInGroup)
			return null;

		groups.put(new TuplesGroup(list1), list1_private);
		groups.put(new TuplesGroup(list0), list0_private);

		return groups;
	}

	public double getLastPrivateSplitEfficiency()
	{
		return lastPrivateSplitEfficiency;
	}

	public String getOutFileName()
	{
		return "KMondrian";
	}

	@Override
	public PrivacyChecker getPrivacyChecker()
	{
		return this.checker;
	}
}
