package edu.emory.mathcs.anonymizers;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import edu.emory.mathcs.Site;
import edu.emory.mathcs.Tuple;
import edu.emory.mathcs.TuplesGroup;
import edu.emory.mathcs.measures.CountSites;
import edu.emory.mathcs.measures.Measure;
import edu.emory.mathcs.privacy.PrivacyChecker;

/**
 * KSiteMondrianSplitterMID + finding median using ID of each site
 */
public class KSiteMondrianSplitterMID implements Anonymizer
{
	private int minNumTuplesInGroup;
	private double alpha = 0.5;
	private double beta = 0.0;
	private PrivacyChecker checker;
	private Measure splittingMeasure;
	private Measure sitesCounter;

	public KSiteMondrianSplitterMID(int minNumTuplesInGroup, double alpha, double beta, Measure splittingMeasure, PrivacyChecker checker)
	{
		this.minNumTuplesInGroup = minNumTuplesInGroup;
		this.alpha = alpha;
		this.beta = beta;
		this.checker = checker;
		this.splittingMeasure = splittingMeasure;
		this.sitesCounter = new CountSites();
		
	}

	@Override
	public String getOutFileName()
	{
		return "KSiteMondrianSplitter";
	}

	@Override
	public void splitGroup(TuplesGroup tuplesToSplit)
	{
		Stack<TuplesGroup> groupsToConsider = new Stack<TuplesGroup>();
		groupsToConsider.push(tuplesToSplit);

		TuplesGroup group;
		while (groupsToConsider.size() > 0)
		{
			group = groupsToConsider.pop();

			if (group.getNumTuples() < 2 * this.minNumTuplesInGroup)
			{
				return;
			}

			int[] rangeLen = new int[group.getNumAttributes()];
			rangeLen[0] = group.getAttributeRange(0).getMax() - group.getAttributeRange(0).getMin();
			int maxRange = rangeLen[0];
			int minRange = maxRange;
			for (int i = 1; i < rangeLen.length; i++)
			{
				rangeLen[i] = group.getAttributeRange(i).getMax() - group.getAttributeRange(i).getMin();
				if (rangeLen[i] > maxRange)
				{
					maxRange = rangeLen[i];
				}
				if (rangeLen[i] < minRange)
				{
					minRange = rangeLen[i];
				}
			}

			Map<Site, List<Tuple>> site2tuples = group.getTuplesSplittedToSites();
			double[] splittingMeasureValues = new double[group.getNumAttributes()];
//			double minMeasureValue = Double.MAX_VALUE;
//			double maxMeasureValue = 0;
//			double[] sitesNum = new double[group.getNumAttributes()];
//			double minNumSites = Double.MAX_VALUE;
//			double maxNumSites = 0.0;
			// chose the best attribute to split
			for (int attributeID = 0; attributeID < group.getNumAttributes(); attributeID++)
			{

				TuplesGroup splittedGroup = this.attributeSplitGroup(group, site2tuples, attributeID);

				if (splittedGroup == null || this.checker.isAnonymized(splittedGroup) == false)
				{
					splittingMeasureValues[attributeID] = Double.NaN;
					continue;
				}

				double leftMeasureValue = this.checker.getPrivacyFitness(splittedGroup.getLeftChild());
				double rightMeasureValue = this.checker.getPrivacyFitness(splittedGroup.getRightChild());
				splittingMeasureValues[attributeID] = Math.min(leftMeasureValue, rightMeasureValue);
//				splittingMeasureValues[attributeID] = this.splittingMeasure.getValue(splittedGroup);
//				sitesNum[attributeID] = this.sitesCounter.getValue(splittedGroup);

//				maxMeasureValue = Math.max(maxMeasureValue, splittingMeasureValues[attributeID]);
//				minMeasureValue = Math.min(minMeasureValue, splittingMeasureValues[attributeID]);
//
//				minNumSites = Math.min(minNumSites, sitesNum[attributeID]);
//				maxNumSites = Math.max(maxNumSites, sitesNum[attributeID]);

			}

			// Now I have everything to find the attribute for split
			int best = -1;
			double bestScore = Double.NaN;
			for (int i = 0; i < splittingMeasureValues.length; i++)
			{
				if (Double.isNaN(splittingMeasureValues[i]))
					continue;
				
//				double factorRange = (minRange == maxRange ? 1.0 : Tools.getProjectedTo01(minRange, maxRange, rangeLen[i]));
//				double factorDistribution = (minMeasureValue == maxMeasureValue ? 1.0 : Tools.getProjectedTo01(minMeasureValue, maxMeasureValue, splittingMeasureValues[i]));
//				double factorSiteDistribution = (minNumSites == maxNumSites ? 1.0 : Tools.getProjectedTo01(minNumSites, maxNumSites, sitesNum[i]));
//				double score = (1.0 - this.alpha - this.beta) * factorRange + this.alpha * factorDistribution + this.beta * factorSiteDistribution;
				double score = splittingMeasureValues[i];

//				if (Double.isNaN(bestScore) || this.splittingMeasure.getBetterScore(score, bestScore) != bestScore)
				if (Double.isNaN(bestScore) || score > bestScore)
				{
					bestScore = score;
					best = i;
				}
			}

			if (best != -1)
			{
				TuplesGroup splittedGroup = this.attributeSplitGroup(group, site2tuples, best);
				if (splittedGroup == null)
					continue;
				if (splittedGroup.getLeftChild() != null)
					groupsToConsider.push(splittedGroup.getLeftChild());
				if (splittedGroup.getRightChild() != null)
					groupsToConsider.push(splittedGroup.getRightChild());
			}
		}

		tuplesToSplit.reduceRedundantGroups();
	}

	protected TuplesGroup attributeSplitGroup(TuplesGroup group, Map<Site, List<Tuple>> site2tuples, int attributeID)
	{
		List<Integer> numTuples = new ArrayList<Integer>(site2tuples.size());
		for (Site s: site2tuples.keySet())
			numTuples.add(s.getID());
		
		if (numTuples.size() == 1)
			return null;
				
		Collections.sort(numTuples);
		double numTuplesOfMedianSite = numTuples.get(numTuples.size() / 2);
		if (numTuples.size() % 2 == 0 && numTuples.size() > 1)
			numTuplesOfMedianSite = (numTuplesOfMedianSite + numTuples.get(numTuples.size() / 2 - 1)) / 2.0;
		
		List<Tuple> smallTuples = new ArrayList<Tuple>();
		List<Tuple> moreTuples = new ArrayList<Tuple>();

		for (Site s: site2tuples.keySet()) {
			List<Tuple> tuples = site2tuples.get(s);
			if (s.getID() < numTuplesOfMedianSite)
				smallTuples.addAll(tuples);
			else if (s.getID() > numTuplesOfMedianSite)
				moreTuples.addAll(tuples);
			else if (smallTuples.size() > moreTuples.size())
				moreTuples.addAll(tuples);
			else
				smallTuples.addAll(tuples);
		}
		
		// speed up for k-anonymity cases
		if (smallTuples.size() < this.minNumTuplesInGroup || moreTuples.size() < this.minNumTuplesInGroup)
			return null;
		
		TuplesGroup smallTuplesGroup = new TuplesGroup(smallTuples);
		TuplesGroup moreTuplesGroup = new TuplesGroup(moreTuples);
		group.setTuplesGroups(smallTuplesGroup, moreTuplesGroup);
		group.clearTuples();
		
		TuplesGroup medianTuples = new TuplesGroup(null, null);
		medianTuples.setTuplesGroups(smallTuplesGroup, moreTuplesGroup);
		return medianTuples;
		
	}

	@Override
	public PrivacyChecker getPrivacyChecker()
	{
		return this.checker;
	}
}
