package edu.emory.mathcs.anonymizers;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;

import org.java.util.concurrent.NotifyingBlockingThreadPoolExecutor;

import edu.emory.mathcs.Site;
import edu.emory.mathcs.Tools;
import edu.emory.mathcs.Tuple;
import edu.emory.mathcs.TuplesGroup;
import edu.emory.mathcs.privacy.AndCheckers;
import edu.emory.mathcs.privacy.K_anonymity;
import edu.emory.mathcs.privacy.PrivacyChecker;

public class ParallelAdaptiveMPrivacySplitterThreads implements Anonymizer
{
	private static enum SPLIT_SITE {SORT_VALUES};
	private int minNumTuplesInGroup;
	private PrivacyChecker checker;
	private long lastPrivateSplitsNum;
	private long lastSplitsNum;
	private final int numThreadsSplitting;
	private long splittingTime;
	private long checkingTime;
	private int runningSplits;
	private Semaphore semaphore;
	
	public ParallelAdaptiveMPrivacySplitterThreads(int minNumTuplesInGroup, PrivacyChecker checker, int numThreadsSplitting)
	{
		this.minNumTuplesInGroup = minNumTuplesInGroup;
		this.checker = checker;
		this.lastPrivateSplitsNum = 0;
		this.lastSplitsNum = 0;
		this.splittingTime = 0;
		this.checkingTime = 0;
		this.numThreadsSplitting = numThreadsSplitting;
		this.runningSplits = 0;
		this.semaphore = new Semaphore(1);
	}

	public synchronized void addSplittingTime(long splittingTime) {
		this.splittingTime += splittingTime;
	}

	public long getSplittingTime() {
		return this.splittingTime;
	}

	public synchronized void addRunningSplit() {
		this.runningSplits++;
	}

	public synchronized void removeRunningSplit() {
		this.runningSplits--;
	}

	public synchronized long getCheckingTime() {
		return this.checkingTime;
	}
	
	public void splitGroup(TuplesGroup group) {
		rawSplitGroup(group);
		group.reduceRedundantGroups();
	}
	
	private void rawSplitGroup(TuplesGroup group) {
		
		long total_time = -Calendar.getInstance().getTimeInMillis();
		final int numAttributes = group.getNumAttributes();
		NotifyingBlockingThreadPoolExecutor executor = new NotifyingBlockingThreadPoolExecutor(this.numThreadsSplitting, 2*this.numThreadsSplitting, 30, TimeUnit.SECONDS);
		LinkedBlockingQueue<TuplesGroup> groupsToConsider = new LinkedBlockingQueue<TuplesGroup>();
		groupsToConsider.add(group);

		do {
//			System.err.println(groupsToConsider.size() + " " + this.getRunnigSplits());
			executor.submit(new GroupSplitTask(groupsToConsider, numAttributes, this, this.minNumTuplesInGroup));
//			new GroupSplitTask(groupsToConsider, numAttributes, this, this.minNumTuplesInGroup).run();
		} while (groupsToConsider.size() > 0 || this.getRunnigSplits() > 0);
		
//		System.err.println(groupsToConsider.size() + " " + this.getRunnigSplits());
		executor.shutdown();
		this.semaphore.release(Integer.MAX_VALUE);
		try
		{
			if (!executor.awaitTermination(10, TimeUnit.SECONDS))
				System.err.println("Timeout!");
		}
		catch (InterruptedException e)
		{
			e.printStackTrace();
		}
		
		total_time += Calendar.getInstance().getTimeInMillis();
		this.addSplittingTime(total_time);
//		System.err.println("MNUM: " + this.lastSplitsNum + " splits, with " + this.lastPrivateSplitsNum + " private.");
	}

	private synchronized int getRunnigSplits()
	{
		return this.runningSplits;
	}

	public double getLastPrivateSplitEfficiency() {
		return (double) this.lastPrivateSplitsNum / (double) this.lastSplitsNum;
	}
	
	public synchronized void incrementLastPrivateSplitsNum() {
		this.lastPrivateSplitsNum++;
	}

	public String getOutFileName()
	{
		return "KAttributeMondrian";
	}

	@Override
	public PrivacyChecker getPrivacyChecker()
	{
		return new AndCheckers.AndCheckersBuilder()
				.addChecker(this.checker, 1.0)
				.addChecker(new K_anonymity(this.minNumTuplesInGroup), 1.0)
				.build("K_mondrian_splitter_checker");
	}
	
	private class GroupSplitTask implements Runnable {
		
		private BlockingQueue<TuplesGroup> groupsToConsider;
		private final int numAttributes;
		private final int minNumTuplesInGroup;
		private final ParallelAdaptiveMPrivacySplitterThreads anonymizer;
		
		public GroupSplitTask(BlockingQueue<TuplesGroup> groupsToConsider, int numAttributes,
				ParallelAdaptiveMPrivacySplitterThreads anonymizer, int minNumTuplesInGroup) {
			this.groupsToConsider = groupsToConsider;
			this.numAttributes = numAttributes;
			this.minNumTuplesInGroup = minNumTuplesInGroup;
			this.anonymizer = anonymizer;
		}
		public void run() {

			this.anonymizer.acuireSemaphore();
			this.anonymizer.addRunningSplit();
//			System.out.print("+");
			
			TuplesGroup group = groupsToConsider.poll();

			if (group.getNumTuples() < 2 * this.minNumTuplesInGroup) {
				this.anonymizer.removeRunningSplit();
//				System.out.println("- (k)");
				return;
			}
			
			PrivacyChecker checker = this.anonymizer.getPrivacyChecker();
			
			// Medians and attribute ranges
			double[] median = new double[numAttributes];
			for (int i = 0; i < median.length; i++)
				median[i] = Tools.findMedian(i, group.getData());
	
			// Possible splits
			List<TuplesGroup[]> groups = new ArrayList<TuplesGroup[]>(numAttributes);
			List<Double> splittingMeasureValues = new ArrayList<Double>();
			TuplesGroup[] splitGroup1 = null;
			List<TuplesGroup[]> splitGroup2 = null;
			for (int i = 0; i < median.length; i++)
			{
				splitGroup1 = group.split(i, median[i], this.minNumTuplesInGroup, null);
				
				this.anonymizer.incrementLastPrivateSplitsNum();
				
				long total_time = -Calendar.getInstance().getTimeInMillis();
				boolean isSplitGroup10Private = checker.isAnonymized(splitGroup1[0]); 
				boolean isSplitGroup11Private = checker.isAnonymized(splitGroup1[1]);
				total_time += Calendar.getInstance().getTimeInMillis();
				this.anonymizer.addCheckingTime(total_time);
				if (isSplitGroup10Private && isSplitGroup11Private)
				{
					this.anonymizer.incrementLastPrivateSplitsNum();
					groups.add(splitGroup1);
				}
			}
	
			splitGroup2 = this.splitSites(group, SPLIT_SITE.SORT_VALUES);
			if (splitGroup2 != null)
			{
				for (int i = 0; i < splitGroup2.size(); i++)
				{
					TuplesGroup[] sg = splitGroup2.get(i);
					if (sg == null)
						continue;
					long total_time = -Calendar.getInstance().getTimeInMillis(); 
					boolean issg0Private = checker.isAnonymized(sg[0]); 
					boolean issg1Private = checker.isAnonymized(sg[1]);
					total_time += Calendar.getInstance().getTimeInMillis(); 
					this.anonymizer.addCheckingTime(total_time);
					if (issg0Private && issg1Private)
					{
						this.anonymizer.incrementLastPrivateSplitsNum();
						groups.add(sg);
					}
				}
			}
	
			for (int i = 0; i < groups.size(); i++)
			{
				TuplesGroup[] sg = groups.get(i);
	
				if (sg == null)
					continue;
	
				double leftMeasureValue = checker.getPrivacyFitness(sg[0]);
				double rightMeasureValue = checker.getPrivacyFitness(sg[1]);
				splittingMeasureValues.add(Math.min(leftMeasureValue, rightMeasureValue));
	
			}
	
			// Now I have rangesLen, maxRange, minNodes, min, max
			int best = -1;
			double bestScore = -1;
			for (int i = 0; i < groups.size(); i++)
			{
				if (groups.get(i) == null)
					continue;
	
				double score = splittingMeasureValues.get(i);
				
				if (score > bestScore)
				{
					bestScore = score;
					best = i;
				}
			}
	
			if (best != -1)
			{
				group.setTuplesGroups(groups.get(best)[0], groups.get(best)[1]);
				if (!this.groupsToConsider.offer(groups.get(best)[0]))
					System.err.println("Unable to add a group to consider!a");
				this.anonymizer.releaseSemaphore();
				if (!this.groupsToConsider.offer(groups.get(best)[1]))
					System.err.println("Unable to add a group to consider!b");
				this.anonymizer.releaseSemaphore();
			}
			this.anonymizer.removeRunningSplit();
//			System.out.println("-");
	
		}
		private List<TuplesGroup[]> splitSites(TuplesGroup group, SPLIT_SITE sortValues)
		{
			TuplesGroup[] groups = new TuplesGroup[2];
			List<Tuple> list0 = new ArrayList<Tuple>();
			List<Tuple> list1 = new ArrayList<Tuple>();
			
			Map<Site, List<Tuple>> site2tuples = group.getTuplesSplittedToSites();
	
			if (sortValues == SPLIT_SITE.SORT_VALUES) {
				
				// get attribute values of first unnecessary tuple in the site 
				final Map<Site, Double> site2Values = new TreeMap<Site, Double>();
				
				if (site2Values.size() == 1)
					return null;
				
				int numAllTuples = 0;
				for (Site site : site2tuples.keySet()) {
					List<Tuple> ts = site2tuples.get(site);
	//				site2Values.put(site, this.checker.getPrivacyFitness(new TuplesGroup(ts, null, null)));
	//				site2Values.put(site, ts.size());
					numAllTuples += ts.size();
				}
				
				// sort sites by value
				List<Site> sites = new ArrayList<Site>(site2tuples.keySet());
	// To test if sorting is needed.
	//			Collections.sort(sites, new Comparator<Site>(){
	//				public int compare(Site o1, Site o2)
	//				{
	//					return (int)Math.signum(site2Values.get(o2) - site2Values.get(o1));
	//				}
	//			});
	//			
				// smart
				int i = 0;
				while (i < sites.size() && (list1.size() < numAllTuples / 2))
				{
					List<Tuple> ts = site2tuples.get(sites.get(i++));
					list1.addAll(ts);
				}
				
				for (; i < sites.size(); i++)
					list0.addAll(site2tuples.get(sites.get(i)));
			}
			
			this.anonymizer.incrementLastPrivateSplitsNum();
			// to speed up for k-anonymity case(s)
			if (list0.size() < this.minNumTuplesInGroup || list1.size() < this.minNumTuplesInGroup)
				return null;
			
			groups[0] = new TuplesGroup(list1);
			groups[1] = new TuplesGroup(list0);
			List<TuplesGroup[]> tg = new ArrayList<TuplesGroup[]>(1);
			tg.add(groups);
			
			return tg;
		}

	}

	public void acuireSemaphore()
	{
		try
		{
//			System.out.println(this.semaphore.toString() + " -1");
			this.semaphore.acquire();
		}
		catch (InterruptedException e)
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	public synchronized void addCheckingTime(long totalTime)
	{
		this.checkingTime += totalTime;
	}

	public void releaseSemaphore() {
//		System.out.println(this.semaphore.toString() + " +1");
		this.semaphore.release();
	}
	
}
