package edu.emory.mathcs.measures.privacy;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import edu.emory.mathcs.Tools;
import edu.emory.mathcs.Tuple;
import edu.emory.mathcs.TuplesGroup;
import edu.emory.mathcs.measures.Measure;

public class DeltaGrowthAverage implements Measure
{
	public enum P_PRIOR {ZERO, EQUAL, ALL_DATA_BASED};
	public static final String NAME = "DeltaGrowthAverage";

	protected P_PRIOR p_prior;

	public DeltaGrowthAverage(P_PRIOR type)
	{
		this.p_prior = type;
	}

	@Override
	public double getBetterScore(double score1, double score2)
	{
		return Math.min(score1, score2);
	}

	@Override
	public double getMaxValue(final Tuple[] tuples)
	{
		if (tuples.length == 0)
			return Double.NaN;

		return 1.0;
	}

	@Override
	public double getMinValue(final Tuple[] tuples)
	{
		if (tuples.length == 0)
			return Double.NaN;

		return 0.0;
	}

	@Override
	public String getName()
	{
		return DeltaGrowthAverage.NAME;
	}

	@Override
	public double getValue(TuplesGroup tuples)
	{

		double numTuples = tuples.getNumTuples();

		if (numTuples == 0.0)
			return Double.NaN;

		Map<Integer, Double> p_prior = Tools.getValueProbabilities(tuples, TuplesGroup.SENSITIVE, false); 
		if (this.p_prior == P_PRIOR.ZERO)
		{
			for (Integer attrValue : p_prior.keySet())
				p_prior.put(attrValue, 0.0);
		}
		
		if (this.p_prior == P_PRIOR.EQUAL)
		{
			for (Integer attrValue : p_prior.keySet())
				p_prior.put(attrValue, 1.0 / p_prior.size());
		}

		if (this.p_prior == P_PRIOR.ALL_DATA_BASED)
			;
		
		double value = this.getValueFromAGroup(tuples, p_prior, tuples);
		return value / (double) tuples.countGroups(false);
	}

	protected double getValueFromAGroup(TuplesGroup tuples, Map<Integer, Double> probs, final TuplesGroup allTuples)
	{

		double groupOverProbability = 0.0;
		if (tuples.hasSubGroups())
		{
			groupOverProbability = this.getCombinedMeasureValue(this
					.getValueFromAGroup(tuples.getLeftChild(), probs, allTuples), this
					.getValueFromAGroup(tuples.getRightChild(), probs, allTuples));
		}
		else
		{
			List<TuplesGroup> intersectingGroups = allTuples.getIntersectingLeafGroups(tuples);
			List<Tuple> ts = new ArrayList<Tuple>();
			
			ts.addAll(tuples.getUncorruptedTuplesAsList());
			for (TuplesGroup g : intersectingGroups)
				ts.addAll(g.getUncorruptedTuplesAsList());
			
			TuplesGroup tg = new TuplesGroup(ts);

			Map<Integer, Integer> freq = tg.getAttributeFrequenciesUncorrupted(TuplesGroup.SENSITIVE);
			int numTuples = tg.getNumTuples();

			for (Integer value : freq.keySet())
			{
				// Double prob = probs.get(value);
				Double prob = probs.get(0);
				if (prob == null)
					continue;

				double overProbability = (double) freq.get(value) / (double) numTuples - prob;
				if (overProbability > 0.0)
					groupOverProbability += overProbability;
			}
		}

		return groupOverProbability;
	}

	@Override
	public int compareTo(Measure o)
	{
		return o.getName().compareTo(this.getName());
	}

	@Override
	public double getCombinedMeasureValue(double leftValue, double rightValue)
	{
		return leftValue + rightValue;
	}
}
