package edu.emory.mathcs.measures.privacy;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import edu.emory.mathcs.Tuple;
import edu.emory.mathcs.TuplesGroup;
import edu.emory.mathcs.measures.Measure;

/**
 * Based on:
 * 		t-Closeness: Privacy Beyond k-Anonymity and l-Diversity
 * by 
 * 		Ninghui Li, Tiancheng Li and Suresh Venkatasubramanian
 * 
 * @note is normalized to number of tuples
 */
public class DeltaGrowthGroup implements Measure {

	public enum P_PRIOR {ZERO, EQUAL, ALL_DATA_BASED};
	public static final String NAME = "DeltaGrowthGroup";
	
	protected P_PRIOR p_prior;
//	protected TuplesGroup alltuples;
	
//	public DeltaGrowthGroup(P_PRIOR type, TuplesGroup alltuples) {
//		this.p_prior = type;
//		this.alltuples = alltuples;
//	}

	public DeltaGrowthGroup(P_PRIOR type) {
		this.p_prior = type;
//		this.alltuples = null;
	}

	@Override
	public double getBetterScore(double score1, double score2) {
		return Math.min(score1, score2);
	}

	@Override
	public double getMaxValue(final Tuple[] tuples) {
		if (tuples.length == 0)
			return Double.NaN;
		
		return 1.0;
	}

	@Override
	public double getMinValue(final Tuple[] tuples) {
		if (tuples.length == 0)
			return Double.NaN;

		return 0.0;
	}

	@Override
	public String getName() {
		return DeltaGrowthGroup.NAME;
	}

	@Override
	public double getValue(TuplesGroup tuples) {
		
		double numTuples = tuples.getNumTuples();
		
		if (numTuples == 0.0)
			return Double.NaN;
		
		Map<Integer, Double> probs = null;
		if (this.p_prior == P_PRIOR.ZERO)
		{
			probs = tuples.getAttributeValueProbabilities(TuplesGroup.SENSITIVE, false);
			for (Integer attrValue : probs.keySet())
				probs.put(attrValue, 0.0);
		}
		
		if (this.p_prior == P_PRIOR.EQUAL)
		{
			probs = tuples.getAttributeValueProbabilities(TuplesGroup.SENSITIVE, false);
			for (Integer attrValue : probs.keySet())
				probs.put(attrValue, 1.0 / probs.size());
		}

		if (this.p_prior == P_PRIOR.ALL_DATA_BASED)
			probs = tuples.getAttributeValueProbabilities(TuplesGroup.SENSITIVE, true); 
		
		double value = this.getValueFromAGroup(tuples, probs, tuples);
		return value;
	}

	protected double getValueFromAGroup(TuplesGroup tuples, Map<Integer, Double> probs, final TuplesGroup allTuples) {
		
		double groupOverProbability = 0.0;
		if (tuples.hasSubGroups())
		{
			groupOverProbability = this.getCombinedMeasureValue(this.getValueFromAGroup(tuples.getLeftChild(), probs, allTuples), this.getValueFromAGroup(tuples.getRightChild(), probs, allTuples));
		}
		else
		{
			List<TuplesGroup> intersectingGroups = allTuples.getIntersectingLeafGroups(tuples);
			List<Tuple> ts = new ArrayList<Tuple>();
			
			ts.addAll(tuples.getUncorruptedTuplesAsList());
			for (TuplesGroup g : intersectingGroups)
				ts.addAll(g.getUncorruptedTuplesAsList());
			
			TuplesGroup tg = new TuplesGroup(ts);
			
			Map<Integer, Integer> freq = tg.getAttributeFrequenciesUncorrupted(TuplesGroup.SENSITIVE);
			int numTuples = tg.getNumTuples();
			
			for (Integer value : freq.keySet())
			{
				Double prob = probs.get(value);
				if (prob == null)
					continue;
				
				double overProbability = (double)freq.get(value) / (double)numTuples - prob;
				if (overProbability > 0.0)
					groupOverProbability += overProbability;
			}
		}
		
		return groupOverProbability;
	}

	@Override
	public int compareTo(Measure o) {
		return o.getName().compareTo(this.getName());
	}

	@Override
	public double getCombinedMeasureValue(double leftValue, double rightValue) {
		return Math.max(leftValue, rightValue);
	}

}
