package edu.emory.mathcs.measures.privacy;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import edu.emory.mathcs.Tuple;
import edu.emory.mathcs.TuplesGroup;
import edu.emory.mathcs.measures.Measure;

/**
 * Change in the probability of disclosing a record.
 *   max_{all q-groups} (max_{sensitive values s in a q-group} (P_{posteriori}(s) - P_{priori}(s)))
 * P_{priori}(s) can be defined as 0, equal for all possible values or based on data sample -> DisclosureProbabilityMaxMax.P_PRIOR
 * @author Slawomir Goryczka
 */
public class DeltaGrowthValue implements Measure {

	public enum P_PRIOR {ZERO, EQUAL, ALL_DATA_BASED};
	public static final String NAME = "DeltaGrowthValue";
	
	protected P_PRIOR p_prior;
	
	public DeltaGrowthValue(P_PRIOR type) 
	{
		this.p_prior = type;
	}
	
	public double getMaxValue(final Tuple[] tuples) {
		
		if (tuples.length == 0)
			return Double.NaN;
		
		return 1.0;
	}

	protected Map<Integer, Integer> getValuesFrequencies(final Tuple[] tuples)
	{
		Map<Integer, Integer> value2freq = new TreeMap<Integer, Integer>();
		
		for (Tuple t : tuples)
		{
			Integer value = t.getData(TuplesGroup.SENSITIVE);
			Integer freq = value2freq.get(value);
			
			if (freq == null)
				freq = 0;
			
			value2freq.put(value, freq+1);
		}
		
		return value2freq;
	}
	
	public double getMinValue(final Tuple[] tuples) {
		
		if (tuples.length == 0)
			return Double.NaN;
		
		return 0.0;
	}

	@Override
	public String getName() {
		return DeltaGrowthValue.NAME;
	}

	@Override
	public double getValue(TuplesGroup tuples) {
		double p_prior = Double.NaN;
		
		if (this.p_prior == P_PRIOR.ZERO)
			p_prior = 0.0;
		
		if (this.p_prior == P_PRIOR.EQUAL)
			p_prior = 1.0 / tuples.getDomain(TuplesGroup.SENSITIVE, false).size();

		// P_prior is computed with considering corrupted tuples (if any)
		if (this.p_prior == P_PRIOR.ALL_DATA_BASED)
			p_prior = this.getProbability(tuples.getTuples());
		
		return this.getPostProbability(tuples, tuples) - p_prior;
	}

	protected double getPostProbability(TuplesGroup tuples, final TuplesGroup allTuples) {
		if (tuples.hasSubGroups())
		{
			return this.getCombinedMeasureValue(this.getPostProbability(tuples.getLeftChild(), allTuples), this.getPostProbability(tuples.getRightChild(), allTuples));
		}
		else
		{	
			// There may be other groups which attribute ranges overlap with this one
			// They should be included to the domain of the probability - since they are
			// indistinguishable based on at least one QID
			List<TuplesGroup> intersectingGroups = allTuples.getIntersectingLeafGroups(tuples);
			List<Tuple> ts = new ArrayList<Tuple>();
			
			ts.addAll(tuples.getUncorruptedTuplesAsList());
			for (TuplesGroup g : intersectingGroups)
				ts.addAll(g.getUncorruptedTuplesAsList());
			
			return this.getProbability(ts.toArray(new Tuple[0]));
		}
	}

	protected double getProbability(final Tuple[] tuples)
	{
		
		if (tuples.length == 0)
			return Double.NaN;

		Map<Integer, Integer> values2freqs = this.getValuesFrequencies(tuples);
		int max = 0;
		int sum = 0;
		for (Integer freq : values2freqs.values())
		{
			max = Math.max(max, freq);
			sum += freq;	
		}
		
		return (double) max / (double) sum;
	}
	
	@Override
	public int compareTo(Measure o) {
		return o.getName().compareTo(this.getName());
	}

	@Override
	public double getBetterScore(double score1, double score2) {
		return Math.min(score1, score2);
	}

	@Override
	public double getCombinedMeasureValue(double leftValue, double rightValue) {
		if (Double.isNaN(leftValue))
			return rightValue;
		if (Double.isNaN(rightValue))
			return leftValue;
		
		return Math.max(leftValue, rightValue);
	}

}
