package edu.emory.mathcs.measures.utility;

import edu.emory.mathcs.Tuple;
import edu.emory.mathcs.TuplesGroup;
import edu.emory.mathcs.measures.Measure;

/**
 * Based on:
 * 		Transforming Data to Satisfy Privacy Constraints
 * by
 * 		Vijay S. Iyengar
 * 
 * Note: LM for all attributes is considered as sum of LM for each attribute.
 */
public class LossMetric implements Measure {

	public static final String NAME = "LM";
	
	@Override
	public double getMaxValue(final Tuple[] tuples) {
		if (tuples.length == 0)
			return Double.NaN;
			
		return tuples[0].getAttributesNumber();
	}

	@Override
	public double getMinValue(final Tuple[] tuples) {
		
		if (tuples.length == 0)
			return Double.NaN;
		
		return 0.0;
	}
	
	@Override
	public String getName() {
		return LossMetric.NAME;
	}

	@Override
	public double getValue(TuplesGroup tuples)
	{
		double lm = 0;
		
		for (int i = 0; i < tuples.getNumAttributes(); i++)
			lm += this.getValue(tuples, i);
		
		return lm;
	}

	protected double getValue(TuplesGroup tuples, int attributeIndex) 
	{
		double rangeLen = tuples.getAttributeRange(attributeIndex).getMax() - tuples.getAttributeRange(attributeIndex).getMin();

		if (rangeLen == 0)
			return 0.0;
		
		double numTuples = tuples.getNumUncorruptedTuples();
		if (numTuples == 0.0)
			return Double.NaN;
		
		double lmValue = this.getLM(tuples, attributeIndex) / (rangeLen * numTuples);
		
		return lmValue;
	}

	protected double getLM(TuplesGroup tuples, int attributeIndex) 
	{
		double value = 0.0;
		
		if (tuples.hasSubGroups())
		{
			value = this.getCombinedMeasureValue(this.getLM(tuples.getLeftChild(), attributeIndex), this.getLM(tuples.getRightChild(), attributeIndex));
		}
		else
		{
			value = tuples.getAttributeRange(attributeIndex).getMax() - tuples.getAttributeRange(attributeIndex).getMin();
			value *= tuples.getNumUncorruptedTuples();
		}
		
		return value;

	}
	
	@Override
	public int compareTo(Measure o) {
		return o.getName().compareTo(this.getName());
	}

	@Override
	public double getBetterScore(double score1, double score2) {
		return Math.min(score1, score2);
	}

	@Override
	public double getCombinedMeasureValue(double leftValue, double rightValue) {
		return leftValue + rightValue;
	}

}
