package edu.emory.mathcs.privacy;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import edu.emory.mathcs.CombinationsGenerator;
import edu.emory.mathcs.CorruptionAttack;
import edu.emory.mathcs.Site;
import edu.emory.mathcs.Tuple;
import edu.emory.mathcs.TuplesGroup;

public class M_privacy_BIN implements PrivacyChecker
{
	private int m;
	private boolean withAdaptiveOrdering;
	private PrivacyChecker checker;
	
	// stats
	private HashMap<Integer, Integer> numAttackers2Freq;
	private HashMap<Integer, Integer> corruptedTuples2Freq;
	private HashMap<Integer, Integer> numDefenders2Freq;
	private HashMap<Integer, Integer> defendedTuples2Freq;
	
	/**
	 * 
	 * @param m
	 * @param sequenceOfAttackingCoalitions sequence of attackers coalition sizes
	 * negative values counts from the end, i.e. -1 is a num of providers.
	 */
	public M_privacy_BIN(PrivacyChecker checker, int m,
			boolean withAdaptiveOrdering, boolean withStats) {
		this.m = m;
		this.checker = checker;
		this.withAdaptiveOrdering = withAdaptiveOrdering;
		if (withStats) {
			this.corruptedTuples2Freq = new HashMap<Integer, Integer>();
			this.numAttackers2Freq = new HashMap<Integer, Integer>();
			this.numDefenders2Freq = new HashMap<Integer, Integer>();
			this.defendedTuples2Freq = new HashMap<Integer, Integer>();
		}
	}
	
	@Override
	public double getPrivacyFitness(TuplesGroup tuples)
	{
		return this.checker.getPrivacyFitness(tuples);
	}

	@Override
	public boolean isAnonymized(TuplesGroup tuples)
	{
		if (tuples.hasSubGroups()){
			return this.isAnonymized(tuples.getLeftChild()) && this.isAnonymized(tuples.getRightChild());
		}
		else{
			Map<Site, List<Tuple>> site2tuples = tuples.getTuplesSplittedToSites();
			if (site2tuples.size() == 1)
				return this.checker.isAnonymized(tuples);
			return this.isAnonymized(tuples, site2tuples);
		}
	}
	
	private boolean isPrivate(TuplesGroup tuples, Set<Site> attackers) {
		
		final int numSites = tuples.getSites().size();
		
		// number of attackers == number of sites in the tuples group
		if (numSites == attackers.size())
			return true;
		
		if (numSites <= 1)
			return this.checker.isAnonymized(tuples);
		
		int numCorruptedTuples = CorruptionAttack.corruptSitesTuples(tuples, attackers);
		
		this.reportCheck(attackers.size(), numCorruptedTuples,
				numSites - attackers.size(),
				tuples.getNumTuples() - numCorruptedTuples);
		
		boolean isAnonymized = this.checker.isAnonymized(tuples);
		
		int numUncorruptedTuples = CorruptionAttack.uncorruptSitesTuples(tuples, attackers);
		
		if (numCorruptedTuples != numUncorruptedTuples)
			throw new IllegalArgumentException("Number of corrupted tuples: " + numCorruptedTuples
					+ ", while number of uncorrupted tuples: " + numUncorruptedTuples);
		
		return isAnonymized;
	}
	
	private void reportCheck(int numAttackers, int corruptedTuples,
			int numDefenders, int defendedTuples) {
		
		if (this.numAttackers2Freq == null)
			return;
		
		Integer freq = this.numAttackers2Freq.get(numAttackers); 
		if (freq == null)
			freq = 0;
		this.numAttackers2Freq.put(numAttackers, freq + 1);

		freq = this.corruptedTuples2Freq.get(corruptedTuples); 
		if (freq == null)
			freq = 0;
		this.corruptedTuples2Freq.put(corruptedTuples, freq + 1);

		freq = this.numDefenders2Freq.get(numDefenders); 
		if (freq == null)
			freq = 0;
		this.numDefenders2Freq.put(numDefenders, freq + 1);

		freq = this.defendedTuples2Freq.get(defendedTuples); 
		if (freq == null)
			freq = 0;
		this.defendedTuples2Freq.put(defendedTuples, freq + 1);
	}
	
	public void resetStats(){
		this.corruptedTuples2Freq.clear();
		this.numAttackers2Freq.clear();
		this.numDefenders2Freq.clear();
		this.defendedTuples2Freq.clear();
	}
	
	private boolean isAnonymized(TuplesGroup tuples, Map<Site, List<Tuple>> site2tuples) {
		List<Site> sites = new ArrayList<Site>(site2tuples.keySet());
		int maxNumAttackers = sites.size() - 1;
		final int m = Math.min(this.m, maxNumAttackers);
		int numAttackers = maxNumAttackers;
		
		CombinationsGenerator<Site> revGenerator = null;
		if (this.withAdaptiveOrdering){
			M_privacy_Sequence.sortSitesForAdaptiveOrdering(sites, site2tuples, this.checker);
			List<Site> revSites = new ArrayList<Site>(sites);
			Collections.reverse(revSites);
			revGenerator = new CombinationsGenerator<Site>(revSites);
		}
			
		CombinationsGenerator<Site> coalitionsGenerator = new CombinationsGenerator<Site>(sites);
		CombinationsGenerator<Site> mCoalitionsChecker = new CombinationsGenerator<Site>(sites);
		if (revGenerator == null)
			revGenerator = coalitionsGenerator;
		mCoalitionsChecker.setCardinality(m);

		while (!mCoalitionsChecker.areThisAndAllNextDecided()){
			int numMoreAttackers = numAttackers;
			int numLessAttackers = m;
			
			// check less attackers case (bottom one)
			coalitionsGenerator.setCardinality(numLessAttackers);
			while (!coalitionsGenerator.hasNext() && !revGenerator.hasNext()) {
				coalitionsGenerator.setCardinality(++numLessAttackers);
				revGenerator.setCardinality(numLessAttackers);
			}

			Set<Site> lessAttackers = (numLessAttackers >= m ? coalitionsGenerator.next() : revGenerator.next());
			boolean isAnonymizedForLessAttackers = this.isPrivate(tuples, lessAttackers);
			if (lessAttackers.size() <= maxNumAttackers 
					&& !isAnonymizedForLessAttackers && lessAttackers.size() <= m)
				return false; // early stop

			// prepare set of more attackers
			Set<Site> moreAttackers = coalitionsGenerator.nextSuperSet(lessAttackers, numMoreAttackers - numLessAttackers);
			while (moreAttackers == null && numAttackers > numLessAttackers) {
				numMoreAttackers = --numAttackers;
				coalitionsGenerator.setCardinality(numMoreAttackers);
				revGenerator.setCardinality(numMoreAttackers);
				moreAttackers = (numMoreAttackers >= m ? coalitionsGenerator.next() : revGenerator.next());
				if (moreAttackers!= null && moreAttackers.equals(lessAttackers)){
					moreAttackers = (numMoreAttackers >= m ? coalitionsGenerator.next() : revGenerator.next());
				}
			}
			if (numMoreAttackers <= numLessAttackers) {
				moreAttackers = lessAttackers;
			}
			
			// Do privacy check for more attackers
			boolean isAnonymizedForMoreAttackers = this.isPrivate(tuples, moreAttackers);
			if (moreAttackers.size() < maxNumAttackers 
					&& !isAnonymizedForMoreAttackers && moreAttackers.size() <= m)
				return false; // early stop
			
			
			int step = 0;
			Set<Site> currentAttackers = null;
			
			// find a point where adding tuples from a source change the privacy
			while (isAnonymizedForLessAttackers != isAnonymizedForMoreAttackers 
					&& Math.abs(numMoreAttackers - numLessAttackers) > 1){

				step = Math.max(1, (numMoreAttackers - numLessAttackers) / 2);
				currentAttackers = (lessAttackers.size() + step >= m ? coalitionsGenerator.nextSuperSet(lessAttackers, step) : revGenerator.nextSuperSet(lessAttackers, step));
				//currentAttackers = coalitionsGenerator.nextSuperSet(lessAttackers, step);
				
				// In case all (numLessAttackers+step)-combinations are checked
				while (currentAttackers == null && numLessAttackers + step < numMoreAttackers) {
					currentAttackers = (lessAttackers.size() + ++step >= m ? coalitionsGenerator.nextSuperSet(lessAttackers, step) : revGenerator.nextSuperSet(lessAttackers, step));
//					currentAttackers = coalitionsGenerator.nextSuperSet(lessAttackers, ++step);
				}
				
				// if all combinations are checked
				if (numLessAttackers + step >= numMoreAttackers){
					break;
				}
				
				boolean isAnonymized = this.isPrivate(tuples, currentAttackers);
				if (currentAttackers.size() < maxNumAttackers 
						&& !isAnonymized && currentAttackers.size() <= m)
					return false; //early stop

				// update lessAttackers or moreAttackers, if needed
				if (isAnonymized) {
					isAnonymizedForLessAttackers = isAnonymized;
					if (isAnonymizedForLessAttackers == isAnonymizedForMoreAttackers)
						break;
					lessAttackers = currentAttackers;
					numLessAttackers = currentAttackers.size();
				}
				else {
					isAnonymizedForMoreAttackers = isAnonymized;
					if (isAnonymizedForLessAttackers == isAnonymizedForMoreAttackers)
						break;
					moreAttackers = currentAttackers;
					numMoreAttackers = currentAttackers.size();
					
					// it won't be private, since for <= m attackers remaining
					// data is not private
					if (numMoreAttackers <= m)
						return false;
				}
				
			}

			// Update combination generators - by forbidding groups of super
			// and sub combinations.
			coalitionsGenerator.forbidThisAndAllSub(lessAttackers);
			revGenerator.forbidThisAndAllSub(lessAttackers);
			mCoalitionsChecker.forbidThisAndAllSub(lessAttackers);
			coalitionsGenerator.forbidThisAndAllSuper(moreAttackers);
			revGenerator.forbidThisAndAllSuper(moreAttackers);
			mCoalitionsChecker.forbidThisAndAllSuper(moreAttackers);

		}
		return true;
	}

	@Override
	public boolean isMonotonic()
	{
		// it's not entirely true, but for now it's enough
		return this.checker.isMonotonic();
	}

}
