package edu.emory.mathcs.privacy;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import edu.emory.mathcs.CombinationsGenerator;
import edu.emory.mathcs.CorruptionAttack;
import edu.emory.mathcs.Site;
import edu.emory.mathcs.Tuple;
import edu.emory.mathcs.TuplesGroup;

public class M_privacy_Sequence implements PrivacyChecker
{
	private int m;
	private boolean withAdaptiveOrdering;
	private boolean withSwitch;
	private PrivacyChecker checker;
	private List<Integer> sequenceOfAttackingCoalitions;
	
	// stats
	private HashMap<Integer, Integer> numAttackers2Freq;
	private HashMap<Integer, Integer> corruptedTuples2Freq;
	private HashMap<Integer, Integer> numDefenders2Freq;
	private HashMap<Integer, Integer> defendedTuples2Freq;

	/**
	 * 
	 * @param m
	 * @param sequenceOfAttackingCoalitions sequence of attackers coalition sizes
	 * negative values counts from the end, i.e. -1 is a num of providers-1.
	 */
	public M_privacy_Sequence(PrivacyChecker checker, int m,
			List<Integer> sequenceOfAttackingCoalitions, boolean withAdaptiveOrdering,
			boolean withSwitch, boolean withStats) {
		this.m = m;
		this.checker = checker;
		this.sequenceOfAttackingCoalitions = sequenceOfAttackingCoalitions;
		this.withAdaptiveOrdering = withAdaptiveOrdering;
		this.withSwitch = withSwitch;
		if (withStats) {
			this.corruptedTuples2Freq = new HashMap<Integer, Integer>();
			this.numAttackers2Freq = new HashMap<Integer, Integer>();
			this.numDefenders2Freq = new HashMap<Integer, Integer>();
			this.defendedTuples2Freq = new HashMap<Integer, Integer>();
		}
	}
	
	@Override
	public double getPrivacyFitness(TuplesGroup tuples)
	{
		return this.checker.getPrivacyFitness(tuples);
	}

	public static void sortSitesForAdaptiveOrdering(List<Site> sites, Map<Site, List<Tuple>>site2tuples, PrivacyChecker checker){
		Map<Site, Double> site2privacyLevel = new HashMap<Site, Double>();
		for (Site s: sites){
			TuplesGroup group = new TuplesGroup(site2tuples.get(s));
			site2privacyLevel.put(s, checker.getPrivacyFitness(group));
		}

		final Map<Site, Double> finalSite2privacyLevel = new HashMap<Site, Double>(site2privacyLevel);
		Collections.sort(sites, new Comparator<Site>() {
			@Override
			public int compare(Site site0, Site site1)
			{
				return -(int)Math.signum(finalSite2privacyLevel.get(site1) - finalSite2privacyLevel.get(site0));				
			}});
	}
	
	@Override
	public boolean isAnonymized(final TuplesGroup tuples)
	{
		if (tuples.hasSubGroups()){
			return this.isAnonymized(tuples.getLeftChild()) && this.isAnonymized(tuples.getRightChild());
		}
		else{
			Map<Site, List<Tuple>> site2tuples = tuples.getTuplesSplittedToSites();
			if (site2tuples.size() == 1)
				return this.checker.isAnonymized(tuples);
			return this.isAnonymized(tuples, site2tuples);		
		}
	}
	
	private boolean isPrivate(TuplesGroup tuples, Set<Site> attackers) {
		
		final int numSites = tuples.getSites().size();
		
		// number of attackers == number of sites in the tuples group
		if (numSites == attackers.size())
			return true;
		
		if (numSites <= 1)
			return this.checker.isAnonymized(tuples);
		
		int numCorruptedTuples = CorruptionAttack.corruptSitesTuples(tuples, attackers);
		
		this.reportCheck(attackers.size(), numCorruptedTuples,
				numSites - attackers.size(),
				tuples.getNumTuples() - numCorruptedTuples);
		
		boolean isAnonymized = this.checker.isAnonymized(tuples);
		
		int numUncorruptedTuples = CorruptionAttack.uncorruptSitesTuples(tuples, attackers);
		
		if (numCorruptedTuples != numUncorruptedTuples)
			throw new IllegalArgumentException("Number of corrupted tuples: " + numCorruptedTuples
					+ ", while number of uncorrupted tuples: " + numUncorruptedTuples);
		
		return isAnonymized;
	}
	
	private boolean isAnonymized(final TuplesGroup tuples, Map<Site, List<Tuple>> site2tuples) {
		List<Site> sites = new ArrayList<Site>(site2tuples.keySet());
		final int m = Math.min(this.m, sites.size() - 1);
		
		CombinationsGenerator<Site> revGenerator = null;
		if (this.withAdaptiveOrdering){
			sortSitesForAdaptiveOrdering(sites, site2tuples, this.checker);
			List<Site> revSites = new ArrayList<Site>(sites);
			Collections.reverse(revSites);
			revGenerator = new CombinationsGenerator<Site>(revSites);
		}
			
		CombinationsGenerator<Site> coalitionsGenerator = new CombinationsGenerator<Site>(sites);
		CombinationsGenerator<Site> mCoalitionsChecker = new CombinationsGenerator<Site>(sites);
		if (revGenerator == null)
			revGenerator = coalitionsGenerator;
		mCoalitionsChecker.setCardinality(m);
		
		for (int cardinality : this.sequenceOfAttackingCoalitions) {
			
			// interpret negative cardinality
			if (cardinality < 0)
				cardinality += sites.size();
			
			// for m > num_sites take m = num_sites - 1 (all but one providers collude)
			cardinality = Math.min(site2tuples.size() - 1, cardinality);
			if (cardinality < 0)
				continue;
			
			// TODO for parallel tests
//			if (cardinality == 0) {
//				int numTuples = tuples.getNumTuples();
//				// update stats if they're recorded
//				this.reportCheck(0, numTuples, sites.size(), numTuples);
//				if (!this.isPrivate(tuples, new TreeSet<Site>())){
//					// no need to do anything else :)
//					return false;
//				}
//				continue;
//			}

			coalitionsGenerator.setCardinality(cardinality);
			revGenerator.setCardinality(cardinality);
			while ((coalitionsGenerator.hasNext() || revGenerator.hasNext()) 
					&& !mCoalitionsChecker.areThisAndAllNextDecided()) {

				// for cardinality >= m use weaker attackers first!
				Set<Site> attackingSites = (cardinality >= m ? coalitionsGenerator.next() : revGenerator.next());
				if (attackingSites == null)
					throw new IllegalArgumentException("Generators are not synchronized.");
				
				// privacy check
				// TODO: consider non-monotonic privacy constraints
				// use 'corrupted' and introduce 'excluded' concepts to
				// TuplesGroup, which needs to be reviewed. 
				if (!this.isPrivate(tuples, attackingSites)){
					coalitionsGenerator.forbidThisAndAllSuper(attackingSites);
					revGenerator.forbidThisAndAllSuper(attackingSites);
					if (cardinality <= m)
						return false; // early stop
					else
						// just to keep generators synchronized, may be skipped
						mCoalitionsChecker.forbidThisAndAllSuper(attackingSites);
					if (this.withSwitch)
						break; // chose next cardinality (next m) to check
				}
				else {
					coalitionsGenerator.forbidThisAndAllSub(attackingSites);
					revGenerator.forbidThisAndAllSub(attackingSites);
					mCoalitionsChecker.forbidThisAndAllSub(attackingSites);
				}
			}
			
			if (mCoalitionsChecker.areThisAndAllNextDecided())
				break;
		}
		
		return true;
	}
	
	private void reportCheck(int numAttackers, int corruptedTuples,
			int numDefenders, int defendedTuples) {
		
		if (this.numAttackers2Freq == null)
			return;
		
		Integer freq = this.numAttackers2Freq.get(numAttackers); 
		if (freq == null)
			freq = 0;
		this.numAttackers2Freq.put(numAttackers, freq + 1);

		freq = this.corruptedTuples2Freq.get(corruptedTuples); 
		if (freq == null)
			freq = 0;
		this.corruptedTuples2Freq.put(corruptedTuples, freq + 1);

		freq = this.numDefenders2Freq.get(numDefenders); 
		if (freq == null)
			freq = 0;
		this.numDefenders2Freq.put(numDefenders, freq + 1);

		freq = this.defendedTuples2Freq.get(defendedTuples); 
		if (freq == null)
			freq = 0;
		this.defendedTuples2Freq.put(defendedTuples, freq + 1);
	}
	
	public void setSequenceOfAttackingCoalitions(final List<Integer> newSequence){
		this.sequenceOfAttackingCoalitions = newSequence;
	}
	
	
	public void resetStats(){
		this.corruptedTuples2Freq.clear();
		this.numAttackers2Freq.clear();
		this.numDefenders2Freq.clear();
		this.defendedTuples2Freq.clear();
	}

	@Override
	public boolean isMonotonic()
	{
		// it's not entirely true, but for now it's enough
		return this.checker.isMonotonic();
	}

}
