package edu.emory.mathcs.privacy;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;

import org.java.util.concurrent.NotifyingBlockingThreadPoolExecutor;

import edu.emory.mathcs.CombinationsGenerator;
import edu.emory.mathcs.Site;
import edu.emory.mathcs.Tuple;
import edu.emory.mathcs.TuplesGroup;

public class M_privacy_Sequence_threads_pipelining implements PrivacyChecker
{
	private int m;
	private boolean withAdaptiveOrdering;
	private PrivacyChecker checker;
	private List<Integer> sequenceOfAttackingCoalitions;
	final private int numThreadsPrivacyChecking;

	/**
	 * 
	 * @param m
	 * @param sequenceOfAttackingCoalitions sequence of attackers coalition sizes
	 * negative values counts from the end, i.e. -1 is a num of providers-1.
	 */
	public M_privacy_Sequence_threads_pipelining(PrivacyChecker checker, int m,
			List<Integer> sequenceOfAttackingCoalitions, boolean withAdaptiveOrdering,
			int numThreadsPrivacyChecking) {
		
		this.numThreadsPrivacyChecking = numThreadsPrivacyChecking;
		this.m = m;
		this.checker = checker;
		this.sequenceOfAttackingCoalitions = sequenceOfAttackingCoalitions;
		this.withAdaptiveOrdering = withAdaptiveOrdering;
	}
	
	@Override
	public double getPrivacyFitness(TuplesGroup tuples)
	{
		return this.checker.getPrivacyFitness(tuples);
	}

	public static void sortSitesForAdaptiveOrdering(List<Site> sites, Map<Site, List<Tuple>>site2tuples, PrivacyChecker checker){
		Map<Site, Double> site2privacyLevel = new HashMap<Site, Double>();
		for (Site s: sites){
			TuplesGroup group = new TuplesGroup(site2tuples.get(s));
			site2privacyLevel.put(s, checker.getPrivacyFitness(group));
		}

		final Map<Site, Double> finalSite2privacyLevel = new HashMap<Site, Double>(site2privacyLevel);
		Collections.sort(sites, new Comparator<Site>() {
			@Override
			public int compare(Site site0, Site site1)
			{
				return -(int)Math.signum(finalSite2privacyLevel.get(site1) - finalSite2privacyLevel.get(site0));				
			}});
	}
	
	@Override
	public boolean isAnonymized(final TuplesGroup tuples)
	{
		boolean isAnonymized = false;
		if (tuples.hasSubGroups()){
		
				isAnonymized = this.isAnonymized(tuples.getLeftChild()) && this.isAnonymized(tuples.getRightChild());
		 
		}
		else{
			Map<Site, List<Tuple>> site2tuples = tuples.getTuplesSplittedToSites();
			if (site2tuples.size() == 1)
				return this.checker.isAnonymized(tuples);
			try
			{
				isAnonymized = this.isAnonymized(tuples, site2tuples);
			}
			catch (InterruptedException e)
			{
				e.printStackTrace();
			}
			catch (ExecutionException e)
			{
				e.printStackTrace();
			}		
		}
		
		return isAnonymized;
	}

	
	@SuppressWarnings("unchecked")
	private boolean isAnonymized(final TuplesGroup tuples, Map<Site, List<Tuple>> site2tuples) throws InterruptedException, ExecutionException {
		List<Site> sites = new ArrayList<Site>(site2tuples.keySet());
		final int m = Math.min(this.m, sites.size() - 1);
		NotifyingBlockingThreadPoolExecutor executor = new NotifyingBlockingThreadPoolExecutor(this.numThreadsPrivacyChecking, 3*this.numThreadsPrivacyChecking, 30, TimeUnit.SECONDS);
		
		CombinationsGenerator<Site> revGenerator = null;
		if (this.withAdaptiveOrdering){
			sortSitesForAdaptiveOrdering(sites, site2tuples, this.checker);
			List<Site> revSites = new ArrayList<Site>(sites);
			Collections.reverse(revSites);
			revGenerator = new CombinationsGenerator<Site>(revSites);
		}
			
		CombinationsGenerator<Site> coalitionsGenerator = new CombinationsGenerator<Site>(sites);
		CombinationsGenerator<Site> mCoalitionsChecker = new CombinationsGenerator<Site>(sites);
		if (revGenerator == null)
			revGenerator = coalitionsGenerator;
		mCoalitionsChecker.setCardinality(m);

		CombinationsGenerator<Site>[] generators = new CombinationsGenerator[3];
		generators[0] = coalitionsGenerator;
		generators[1] = revGenerator;
		generators[2] = mCoalitionsChecker;
		
		for (int cardinality : this.sequenceOfAttackingCoalitions) {
			
			// interpret negative cardinality
			if (cardinality < 0)
				cardinality += sites.size();
			
			// for m > num_sites take m = num_sites - 1 (all but one providers collude)
			cardinality = Math.min(site2tuples.size() - 1, cardinality);
			if (cardinality < 0)
				continue;
			
			coalitionsGenerator.setCardinality(cardinality);
			revGenerator.setCardinality(cardinality);
			CombinationsGenerator<Site> generator = (cardinality >= m ? coalitionsGenerator : revGenerator);
			while ((coalitionsGenerator.hasNext() || revGenerator.hasNext()) 
					&& !mCoalitionsChecker.areThisAndAllNextDecided()) {

				executor.submit(new IsPrivate(tuples, this.checker, generator, generators, m));
//				new IsPrivate(tuples, this.checker, generator, generators, m).run();
			}
			
//			executor.awaitTermination(30, TimeUnit.SECONDS);
			if (mCoalitionsChecker.areThisAndAllNextDecided())
				break;
		}
		executor.shutdown();
		executor.awaitTermination(30, TimeUnit.SECONDS);
		
		if (mCoalitionsChecker.areAllAllCombinationsForbidden())
			return false;
		
		return true;
	}
	
	public void setSequenceOfAttackingCoalitions(final List<Integer> newSequence){
		this.sequenceOfAttackingCoalitions = newSequence;
	}

	@Override
	public boolean isMonotonic()
	{
		// it's not entirely true, but for now it's enough
		return this.checker.isMonotonic();
	}
	
	private class IsPrivate implements Runnable {
		final private TuplesGroup tuples;
		final private Set<Site> attackers;
		final private PrivacyChecker checker;
		final private CombinationsGenerator<Site>[] generators;
		final int m;
		
		public IsPrivate(TuplesGroup tuples, PrivacyChecker checker, CombinationsGenerator<Site> generator, CombinationsGenerator<Site>[] generators, int m)
		{
			this.tuples = tuples;
			this.checker = checker;
			this.attackers = generator.next();
			this.generators = generators;
			this.m = m;
		}

		public void run()
		{
//			System.out.print("+");
			final int numSites = tuples.getSites().size();
			
			// number of attackers == number of sites in the tuples group
			if (numSites == this.attackers.size()) {
				this.generators[0].forbidThisAndAllSub(this.attackers);
				this.generators[1].forbidThisAndAllSub(this.attackers);
				this.generators[2].forbidThisAndAllSub(this.attackers);
				return;
			}
			
			boolean isAnonymized = false;
			if (numSites <= 1) {
				isAnonymized = this.checker.isAnonymized(tuples);
			} else {
				Set<Tuple> uncorruptedTuples = new HashSet<Tuple>();
				for (Tuple t: tuples.getTuples()) {
					if (this.attackers.contains(t.getSite()) == false)
						uncorruptedTuples.add(t);
				}
				TuplesGroup group = new TuplesGroup(uncorruptedTuples);
				isAnonymized = this.checker.isAnonymized(group);
			}
			
			if (!isAnonymized) {
				this.generators[0].forbidThisAndAllSuper(this.attackers);
				this.generators[1].forbidThisAndAllSuper(this.attackers);
				this.generators[2].forbidThisAndAllSuper(this.attackers);
				if (this.generators[0].getCardinality() <= m) {
					this.generators[2].forbidAllCombinations();
					return; // early stop
				}
			} else {
				this.generators[0].forbidThisAndAllSub(this.attackers);
				this.generators[1].forbidThisAndAllSub(this.attackers);
				this.generators[2].forbidThisAndAllSub(this.attackers);
			}

//			System.out.println("-");
		}
	}
}
